'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME, PARAMETER_CONSTANT_CASEID_KEY, PARAMETER_CONSTANT_TIMESTAMP_KEY
from pm4py.algo.filtering.common.timestamp.timestamp_common import get_dt_from_string
from pm4py.util.xes_constants import DEFAULT_TIMESTAMP_KEY
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
import polars as pl
import datetime


class Parameters(Enum):
    TIMESTAMP_KEY = PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY


def filter_traces_contained(
    df: pl.LazyFrame,
    dt1: Union[str, datetime.datetime],
    dt2: Union[str, datetime.datetime],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Get traces that are contained in the given interval

    Parameters
    ----------
    df
        Polars LazyFrame
    dt1
        Lower bound to the interval (possibly expressed as string, but automatically converted)
    dt2
        Upper bound to the interval (possibly expressed as string, but automatically converted)
    parameters
        Possible parameters of the algorithm, including:
            Parameters.TIMESTAMP_KEY -> Attribute to use as timestamp
            Parameters.CASE_ID_KEY -> Column that contains the case ID

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    dt1 = get_dt_from_string(dt1)
    dt2 = get_dt_from_string(dt2)

    # Group by case and get first and last timestamp
    case_boundaries = (
        df.group_by(case_id_glue)
        .agg([
            pl.col(timestamp_key).min().alias("first_timestamp"),
            pl.col(timestamp_key).max().alias("last_timestamp")
        ])
    )
    
    # Filter cases where both first and last timestamp are within bounds
    valid_cases = case_boundaries.filter(
        (pl.col("first_timestamp") >= dt1) &
        (pl.col("last_timestamp") <= dt2)
    ).select(case_id_glue)
    
    # Keep only events from valid cases
    ret = df.join(valid_cases, on=case_id_glue, how="inner")
    
    return ret


def filter_traces_intersecting(
    df: pl.LazyFrame,
    dt1: Union[str, datetime.datetime],
    dt2: Union[str, datetime.datetime],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter traces intersecting the given interval

    Parameters
    ----------
    df
        Polars LazyFrame
    dt1
        Lower bound to the interval (possibly expressed as string, but automatically converted)
    dt2
        Upper bound to the interval (possibly expressed as string, but automatically converted)
    parameters
        Possible parameters of the algorithm, including:
            Parameters.TIMESTAMP_KEY -> Attribute to use as timestamp
            Parameters.CASE_ID_KEY -> Column that contains the case ID

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    dt1 = get_dt_from_string(dt1)
    dt2 = get_dt_from_string(dt2)
    
    # Group by case and get first and last timestamp
    case_boundaries = (
        df.group_by(case_id_glue)
        .agg([
            pl.col(timestamp_key).min().alias("first_timestamp"),
            pl.col(timestamp_key).max().alias("last_timestamp")
        ])
    )
    
    # A trace intersects if:
    # 1. It has an event within the interval (first > dt1 and first < dt2) OR
    # 2. It has an event within the interval (last > dt1 and last < dt2) OR
    # 3. It spans the entire interval (first < dt1 and last > dt2)
    valid_cases = case_boundaries.filter(
        ((pl.col("first_timestamp") > dt1) & (pl.col("first_timestamp") < dt2)) |
        ((pl.col("last_timestamp") > dt1) & (pl.col("last_timestamp") < dt2)) |
        ((pl.col("first_timestamp") < dt1) & (pl.col("last_timestamp") > dt2))
    ).select(case_id_glue)
    
    # Keep only events from valid cases
    ret = df.join(valid_cases, on=case_id_glue, how="inner")
    
    return ret


def apply_events(
    df: pl.LazyFrame,
    dt1: Union[str, datetime.datetime],
    dt2: Union[str, datetime.datetime],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Get a new log containing all the events contained in the given interval

    Parameters
    ----------
    df
        Polars LazyFrame
    dt1
        Lower bound to the interval (possibly expressed as string, but automatically converted)
    dt2
        Upper bound to the interval (possibly expressed as string, but automatically converted)
    parameters
        Possible parameters of the algorithm, including:
            Parameters.TIMESTAMP_KEY -> Attribute to use as timestamp

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    dt1 = get_dt_from_string(dt1)
    dt2 = get_dt_from_string(dt2)

    # Filter events directly
    ret = df.filter(
        (pl.col(timestamp_key) >= dt1) &
        (pl.col(timestamp_key) <= dt2)
    )

    return ret


def filter_traces_attribute_in_timeframe(
    df: pl.LazyFrame,
    attribute: str,
    attribute_value: str,
    dt1: Union[str, datetime.datetime],
    dt2: Union[str, datetime.datetime],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Get a new log containing all the traces that have an event in the given interval with the specified attribute value

    Parameters
    -----------
    df
        LazyFrame
    attribute
        The attribute to filter on
    attribute_value
        The attribute value to filter on
    dt1
        Lower bound to the interval
    dt2
        Upper bound to the interval
    parameters
        Possible parameters of the algorithm, including:
            Parameters.TIMESTAMP_KEY -> Attribute to use as timestamp
            Parameters.CASE_ID_KEY -> Column that contains the case ID

    Returns
    ------------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, DEFAULT_TIMESTAMP_KEY
    )
    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )

    dt1 = get_dt_from_string(dt1)
    dt2 = get_dt_from_string(dt2)

    # Find cases that have the attribute value in the timeframe
    valid_cases = (
        df.filter(
            (pl.col(attribute) == attribute_value) &
            (pl.col(timestamp_key) >= dt1) &
            (pl.col(timestamp_key) <= dt2)
        )
        .select(case_id_glue)
        .unique()
    )

    # Keep all events from those cases
    filtered = df.join(valid_cases, on=case_id_glue, how="inner")

    return filtered


def apply(df, parameters=None):
    del df
    del parameters
    raise Exception("apply method not available for timestamp filter")


def apply_auto_filter(df, parameters=None):
    del df
    del parameters
    raise Exception(
        "apply_auto_filter method not available for timestamp filter"
    )


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Union, Tuple, Set


class Parameters(Enum):
    THETA = "theta"
    ALPHA = "alpha"
    BETA = "beta"
    GAMMA = "gamma"


def apply(real: Union[OCEL,
                      Tuple[Set[str],
                            Dict[Tuple[str,
                                       str,
                                       str],
                                 int]]],
          normative: Tuple[Set[str],
                           Dict[Tuple[str,
                                str,
                                str],
                                int]],
          parameters: Optional[Dict[Any,
                                    Any]] = None) -> Dict[str,
                                                          Any]:
    """
    Applies OTG-based conformance checking between a 'real' object (OCEL or OTG) and a 'normative' OTG.

    Published in: https://publications.rwth-aachen.de/record/1014107

    Parameters
    -----------------
    real
        Real object (OCEL or OTG)
    normative
        Normative OTG
    parameters
        Variant-specific parameters:
        - Parameters.THETA
        - Parameters.ALPHA
        - Parameters.BETA
        - Parameters.GAMMA

    Returns
    -----------------
    conf_diagn
        Diagnostics dictionary
    """
    if parameters is None:
        parameters = {}

    theta = exec_utils.get_param_value(Parameters.THETA, parameters, None)
    alpha = exec_utils.get_param_value(Parameters.ALPHA, parameters, 1)
    beta = exec_utils.get_param_value(Parameters.BETA, parameters, 1)
    gamma = exec_utils.get_param_value(Parameters.GAMMA, parameters, 1)

    if isinstance(real, OCEL):
        from pm4py.algo.discovery.ocel.otg import algorithm as otg_discovery
        real = otg_discovery.apply(real, parameters=parameters)

    return conformance_checking_multigraph(
        real,
        normative,
        theta=theta,
        alpha=alpha,
        beta=beta,
        gamma=gamma)


def conformance_checking_multigraph(
        discovered_otg,
        normative_otg,
        theta=None,
        alpha=1,
        beta=1,
        gamma=1):
    object_types_L, edges_L = discovered_otg
    object_types_M, edges_M = normative_otg

    # Default thresholds for each relationship type if not provided
    if theta is None:
        theta = {
            'object_interaction': 0.2,
            'object_descendants': 0.2,
            'object_inheritance': 0.2,
            'object_cobirth': 0.2,
            'object_codeath': 0.2
        }

    # Object Type Conformance
    missing_object_types = object_types_M - object_types_L
    additional_object_types = object_types_L - object_types_M

    # Edge Conformance
    missing_edges = set(edges_M.keys()) - set(edges_L.keys())
    additional_edges = set(edges_L.keys()) - set(edges_M.keys())

    # Edge Frequency Conformance
    non_conforming_edges = {}
    for edge in set(edges_M.keys()) & set(edges_L.keys()):
        f_M = edges_M[edge]
        f_L = edges_L[edge]
        sigma = edge[1]  # Relationship type
        delta = abs(f_L - f_M) / f_M if f_M != 0 else float('inf')
        if delta > theta.get(sigma, 0.2):
            non_conforming_edges[edge] = delta

    # Fitness Value Calculation
    total_object_types = len(object_types_M)
    total_edges = len(edges_M)
    total_edge_checks = len(edges_M)
    N = alpha * total_object_types + beta * total_edges + gamma * total_edge_checks
    fitness = 1 - (
        alpha * len(missing_object_types) +
        beta * len(missing_edges) +
        gamma * len(non_conforming_edges)
    ) / N

    # Prepare the result
    result = {
        'missing_object_types': missing_object_types,
        'additional_object_types': additional_object_types,
        'missing_edges': missing_edges,
        'additional_edges': additional_edges,
        'non_conforming_edges': non_conforming_edges,
        'fitness': fitness
    }
    return result



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from copy import copy
from enum import Enum
from typing import Optional, Dict, Any, Tuple

from pm4py.objects.dfg.utils import dfg_utils
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils import petri_utils
from pm4py.util import exec_utils, constants


class Parameters(Enum):
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    PARAM_ARTIFICIAL_START_ACTIVITY = constants.PARAM_ARTIFICIAL_START_ACTIVITY
    PARAM_ARTIFICIAL_END_ACTIVITY = constants.PARAM_ARTIFICIAL_END_ACTIVITY


def apply(
    dfg: Dict[Tuple[str, str], int],
    parameters: Optional[Dict[Any, Any]] = None,
):
    """
    Applies the DFG mining on a given object (if it is a Pandas dataframe or a log, the DFG is calculated)

    Parameters
    -------------
    dfg
        Object (DFG) (if it is a Pandas dataframe or a log, the DFG is calculated)
    parameters
        Parameters:
        - Parameters.START_ACTIVITIES: the start activities of the DFG
        - Parameters.END_ACTIVITIES: the end activities of the DFG

    Returns
    -------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    if parameters is None:
        parameters = {}

    start_activities = exec_utils.get_param_value(
        Parameters.START_ACTIVITIES,
        parameters,
        {x: 1 for x in dfg_utils.infer_start_activities(dfg)},
    )
    end_activities = exec_utils.get_param_value(
        Parameters.END_ACTIVITIES,
        parameters,
        {x: 1 for x in dfg_utils.infer_end_activities(dfg)},
    )
    artificial_start_activity = exec_utils.get_param_value(
        Parameters.PARAM_ARTIFICIAL_START_ACTIVITY,
        parameters,
        constants.DEFAULT_ARTIFICIAL_START_ACTIVITY,
    )
    artificial_end_activity = exec_utils.get_param_value(
        Parameters.PARAM_ARTIFICIAL_END_ACTIVITY,
        parameters,
        constants.DEFAULT_ARTIFICIAL_END_ACTIVITY,
    )

    enriched_dfg = copy(dfg)
    for act in start_activities:
        enriched_dfg[(artificial_start_activity, act)] = start_activities[act]
    for act in end_activities:
        enriched_dfg[(act, artificial_end_activity)] = end_activities[act]
    activities = set(x[1] for x in enriched_dfg).union(
        set(x[0] for x in enriched_dfg)
    )
    net = PetriNet("")
    im = Marking()
    fm = Marking()
    left_places = {}
    transes = {}
    right_places = {}
    for act in activities:
        pl1 = PetriNet.Place("source_" + act)
        pl2 = PetriNet.Place("sink_" + act)
        trans = PetriNet.Transition("trans_" + act, act)
        if act in [artificial_start_activity, artificial_end_activity]:
            trans.label = None
        net.places.add(pl1)
        net.places.add(pl2)
        net.transitions.add(trans)
        petri_utils.add_arc_from_to(pl1, trans, net)
        petri_utils.add_arc_from_to(trans, pl2, net)
        left_places[act] = pl1
        right_places[act] = pl2
        transes[act] = trans
    for arc in enriched_dfg:
        hidden = PetriNet.Transition(arc[0] + "_" + arc[1], None)
        net.transitions.add(hidden)
        petri_utils.add_arc_from_to(right_places[arc[0]], hidden, net)
        petri_utils.add_arc_from_to(hidden, left_places[arc[1]], net)
    im[left_places[artificial_start_activity]] = 1
    fm[right_places[artificial_end_activity]] = 1

    return net, im, fm



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union

from pm4py.algo.conformance.alignments.edit_distance.variants import (
    edit_distance,
)
from pm4py.objects.log.obj import EventLog
from pm4py.util import exec_utils
from pm4py.util import typing
import pandas as pd


class Variants(Enum):
    EDIT_DISTANCE = edit_distance


def apply(
    log1: Union[EventLog, pd.DataFrame],
    log2: Union[EventLog, pd.DataFrame],
    variant=Variants.EDIT_DISTANCE,
    parameters: Optional[Dict[Any, Any]] = None,
) -> typing.ListAlignments:
    """
    Aligns each trace of the first log against the second log

    Parameters
    --------------
    log1
        First log
    log2
        Second log
    variant
        Variant of the algorithm, possible values:
        - Variants.EDIT_DISTANCE: minimizes the edit distance
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    aligned_traces
        List that contains, for each trace of the first log, the corresponding alignment
    """
    return exec_utils.get_variant(variant).apply(
        log1, log2, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import re


ref_stri_1 = "ordabcchr"
ref_stri_2 = "ordabc"

re1 = re.compile(r"([^a-zA-Z0-9]+)")
re2 = re.compile(ref_stri_1)


def mask_non_alphanumeric(stri):
    stri_split = re1.split(stri)
    ret = []
    for el in stri_split:
        for char in el:
            if (
                char.isalnum()
                or char == " "
                or char
                in [
                    "(",
                    ")",
                    "*",
                    ".",
                    ",",
                    "'",
                    '"',
                    "=",
                    "<",
                    ">",
                    "_",
                    "+",
                    "-",
                    "!",
                ]
            ):
                ret.append(char)
            else:
                ret.append(
                    ref_stri_1 + ref_stri_2 + str(ord(char)) + ref_stri_1
                )
    return "".join(ret)


def restore_non_alphanumeric(stri):
    stri_split = re2.split(stri)
    ret = []
    for el in stri_split:
        if el.startswith(ref_stri_2):
            ret.append(chr(int(el[len(ref_stri_2):])))
        else:
            ret.append(el)
    return "".join(ret)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Any

from pm4py.algo.reduction.process_tree.variants import tree_tr_based
from pm4py.util import exec_utils


class Variants(Enum):
    TREE_TR_BASED = tree_tr_based


def apply(*args, **kwargs) -> Any:
    """
    Apply a reduction algorithm to a PM4Py object

    Parameters
    ---------------
    args
        Arguments of the reduction algorithm
    kwargs
        Keyword arguments of the reduction algorithm (including the variant, that is an item of the Variants enum)

    Returns
    ---------------
    reduced_obj
        Reduced object
    """
    variant = kwargs["variant"] if "variant" in kwargs else None
    if variant is None:
        raise Exception(
            "please specify the variant of the reduction to be used."
        )
    return exec_utils.get_variant(variant).apply(*args, **kwargs)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
# XES TAGS
TAG_BOOLEAN = "boolean"
TAG_CLASSIFIER = "classifier"
TAG_DATE = "date"
TAG_EVENT = "event"
TAG_EXTENSION = "extension"
TAG_FLOAT = "float"
TAG_GLOBAL = "global"
TAG_ID = "id"
TAG_INT = "int"
TAG_LIST = "list"
TAG_LOG = "log"
TAG_STRING = "string"
TAG_TRACE = "trace"
TAG_VALUES = "values"

# XES 2.0 tags
TAG_CONTAINER = "container"

# XES/INTERNAL KEYS
KEY_CHILDREN = "children"
KEY_KEY = "key"
KEY_KEYS = "keys"
KEY_NAME = "name"
KEY_PREFIX = "prefix"
KEY_SCOPE = "scope"
KEY_URI = "uri"
KEY_VALUE = "value"

DEFAULT_NAME_KEY = "concept:name"
DEFAULT_INSTANCE_KEY = "concept:instance"
DEFAULT_TIMESTAMP_KEY = "time:timestamp"
DEFAULT_START_TIMESTAMP_KEY = "start_timestamp"
DEFAULT_TRACEID_KEY = "concept:name"
DEFAULT_RESOURCE_KEY = "org:resource"
DEFAULT_TRANSITION_KEY = "lifecycle:transition"
DEFAULT_GROUP_KEY = "org:group"



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from typing import Union, Tuple, Any, Optional, Counter as TCounter

import numpy as np
import pandas as pd

from pm4py.objects.dfg.obj import DFG
from pm4py.objects.log.obj import EventLog
from pm4py.util.compression.dtypes import UCL, MCL, UVCL


def is_polars_lazyframe(df: Any) -> bool:
    """Return True if the provided dataframe is a Polars LazyFrame."""
    df_type = str(type(df)).lower()
    return "polars" in df_type and "lazyframe" in df_type


def project_univariate(
    log: Union[EventLog, pd.DataFrame],
    key: str = "concept:name",
    df_glue: str = "case:concept:name",
    df_sorting_criterion_key="time:timestamp",
) -> Optional[UCL]:
    """
    Projects an event log to a univariate list of values
    For example, an event log of the form [[('concept:name':A,'k1':v1,'k2':v2),('concept:name':B,'k1':v3,'k2':v4),...],...]
    is converted to [['A','B',...],...]

    The method returns the compressed log

    :rtype: ``UCL``
    :param log: log to compress (either EventLog or Dataframe)
    :param key: key to use for compression
    :param df_glue: key to use for combining events into traces when the input is a dataframe.
    :param df_sorting_criterion_key: key to use as a sorting criterion for traces (typically timestamps)
    """
    if type(log) is EventLog:
        return [[e[key] for e in t] for t in log]
    else:
        if is_polars_lazyframe(log):
            import polars as pl  # type: ignore[import-untyped]

            log = log.select(
                pl.col(df_glue), pl.col(key), pl.col(df_sorting_criterion_key)
            )
            log = log.sort([df_glue, df_sorting_criterion_key])

            grouped = (
                log.group_by(df_glue, maintain_order=True)
                .agg(pl.col(key).alias("__acts"))
                .collect()
            )

            return grouped["__acts"].to_list()
        else:
            # Pandas dataframe part
            log = log.loc[:, [key, df_glue, df_sorting_criterion_key]]

            cl = list()
            log = log.sort_values(by=[df_glue, df_sorting_criterion_key])
            values = log[key].to_numpy().tolist()
            distinct_ids, start_indexes, case_sizes = np.unique(
                log[df_glue].to_numpy(), return_index=True, return_counts=True
            )
            for i in range(len(distinct_ids)):
                cl.append(
                    values[start_indexes[i]: start_indexes[i] + case_sizes[i]]
                )
            return cl
    return None


def discover_dfg(log: Union[UCL, MCL], index: int = 0) -> DFG:
    """
    Discover a DFG object from a compressed event log (either univariate or multivariate)
    The DFG object represents a counter of integer pairs

    :rtype: ``Counter[Tuple[int, int]]``
    :param log: compressed event log (either uni or multivariate)
    :param indes: index to use for dfg discovery in case of using an multivariate log
    """
    log = _map_log_to_single_index(log, index)
    dfg = DFG()
    [
        dfg.graph.update([(t[i], t[i + 1])])
        for t in log
        for i in range(0, len(t) - 1)
        if len(t)
    ]
    dfg.start_activities.update(get_start_activities(log, index))
    dfg.end_activities.update(get_end_activities(log, index))
    return dfg


def discover_dfg_uvcl(log: UVCL) -> DFG:
    dfg = DFG()
    [
        dfg.graph.update({(t[i], t[i + 1]): log[t]})
        for t in log
        for i in range(0, len(t) - 1)
        if len(t)
    ]
    for a in get_alphabet(log):
        for t in log:
            if len(t) > 0:
                if t[0] == a:
                    dfg.start_activities.update({a: log[t]})
                if t[len(t) - 1] == a:
                    dfg.end_activities.update({a: log[t]})
    return dfg


def get_start_activities(
    log: Union[UCL, MCL, UVCL], index: int = 0
) -> TCounter[Any]:
    log = _map_log_to_single_index(log, index)
    starts = Counter()
    starts.update(map(lambda t: t[0], filter(lambda t: len(t) > 0, log)))
    return starts


def get_end_activities(
    log: Union[UCL, MCL, UVCL], index: int = 0
) -> TCounter[Any]:
    log = _map_log_to_single_index(log, index)
    ends = Counter()
    ends.update(
        map(lambda t: t[len(t) - 1], filter(lambda t: len(t) > 0, log))
    )
    return ends


def get_alphabet(log: Union[UCL, MCL, UVCL], index: int = 0):
    log = _map_log_to_single_index(log, index)
    sorted_set = sorted(set([e for t in log for e in t]))
    return sorted_set


def get_variants(log: Union[UCL, MCL], index: int = 0) -> UVCL:
    log = _map_log_to_single_index(log, index)
    return Counter(map(lambda t: tuple(t), log))


def _map_log_to_single_index(log: Union[UCL, MCL, UVCL], i: int):
    return (
        [list(map(lambda v: v[i], t)) for t in log]
        if type(log) is MCL
        else log
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import importlib.util
import sys
import time
from copy import copy

from pm4py.algo.conformance.alignments.petri_net.variants import (
    state_equation_a_star,
)
from pm4py.objects.log import obj as log_implementation
from pm4py.objects.log.obj import Trace
from pm4py.objects.log.util.xes import DEFAULT_NAME_KEY
from pm4py.objects.petri_net.utils import (
    align_utils as utils,
    decomposition as decomp_utils,
)
from pm4py.statistics.variants.log import get as variants_module
from pm4py.util import exec_utils
from pm4py.util import variants_util

from enum import Enum
from pm4py.util import constants, nx_utils, thread_utils

from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import typing
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    BEST_WORST_COST = "best_worst_cost"
    PARAM_TRACE_COST_FUNCTION = "trace_cost_function"
    ICACHE = "icache"
    MCACHE = "mcache"
    PARAM_THRESHOLD_BORDER_AGREEMENT = "thresh_border_agreement"
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"
    PARAM_MODEL_COST_FUNCTION = "model_cost_function"
    PARAM_SYNC_COST_FUNCTION = "sync_cost_function"
    PARAM_TRACE_NET_COSTS = "trace_net_costs"
    PARAM_MAX_ALIGN_TIME = "max_align_time"
    PARAM_MAX_ALIGN_TIME_TRACE = "max_align_time_trace"
    SHOW_PROGRESS_BAR = "show_progress_bar"


def get_best_worst_cost(
        petri_net, initial_marking, final_marking, parameters=None
):
    if parameters is None:
        parameters = {}

    trace = log_implementation.Trace()

    best_worst, cf = align(
        trace, petri_net, initial_marking, final_marking, parameters=parameters
    )

    best_worst_cost = (
        sum(cf[x] for x in best_worst["alignment"])
        // utils.STD_MODEL_LOG_MOVE_COST
        if best_worst
        else 0
    )

    return best_worst_cost


def apply_from_variants_list_petri_string(
        var_list, petri_net_string, parameters=None
):
    if parameters is None:
        parameters = {}

    from pm4py.objects.petri_net.importer.variants import (
        pnml as petri_importer,
    )

    petri_net, initial_marking, final_marking = (
        petri_importer.import_petri_from_string(petri_net_string)
    )

    res = apply_from_variants_list(
        var_list,
        petri_net,
        initial_marking,
        final_marking,
        parameters=parameters,
    )
    return res


def apply_from_variants_list(
        var_list, petri_net, initial_marking, final_marking, parameters=None
):
    """
    Apply the alignments from the specification of a list of variants in the log

    Parameters
    -------------
    var_list
        List of variants (for each item, the first entry is the variant itself, the second entry may be the number of cases)
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters of the algorithm (same as 'apply' method, plus 'variant_delimiter' that is , by default)

    Returns
    --------------
    dictio_alignments
        Dictionary that assigns to each variant its alignment
    """
    if parameters is None:
        parameters = {}

    # Create traces directly without repeatedly appending to log
    traces = [variants_util.variant_to_trace(varitem[0], parameters=parameters) for varitem in var_list]
    log = log_implementation.EventLog(traces)

    alignment = apply(log, petri_net, initial_marking, final_marking, parameters=parameters)

    # Create dictionary of alignments more efficiently
    dictio_alignments = {var_list[i][0]: alignment[i] for i in range(len(var_list))}

    return dictio_alignments


def apply(
        log: EventLog,
        net: PetriNet,
        im: Marking,
        fm: Marking,
        parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> typing.ListAlignments:
    """
    Apply the recomposition alignment approach
    to a log and a Petri net performing decomposition

    Parameters
    --------------
    log
        Event log
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    parameters
        Parameters of the algorithm

    Returns
    --------------
    aligned_traces
        For each trace, return its alignment
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(
        log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters
    )

    best_worst_cost = get_best_worst_cost(net, im, fm, parameters=parameters)
    parameters[Parameters.BEST_WORST_COST] = best_worst_cost

    list_nets = decomp_utils.decompose(net, im, fm)

    return apply_log(log, list_nets, parameters=parameters)


def apply_log(log, list_nets, parameters=None):
    """
    Apply the recomposition alignment approach
    to a log and a decomposed Petri net
    """
    if parameters is None:
        parameters = {}

    show_progress_bar = exec_utils.get_param_value(
        Parameters.SHOW_PROGRESS_BAR, parameters, constants.SHOW_PROGRESS_BAR
    )

    # Use efficient caches
    icache = exec_utils.get_param_value(Parameters.ICACHE, parameters, dict())
    mcache = exec_utils.get_param_value(Parameters.MCACHE, parameters, dict())

    parameters[Parameters.ICACHE] = icache
    parameters[Parameters.MCACHE] = mcache

    # Get variants more efficiently
    variants_idxs = variants_module.get_variants_from_log_trace_idx(
        log, parameters=parameters
    )

    progress = None
    if importlib.util.find_spec("tqdm") and show_progress_bar:
        from tqdm.auto import tqdm
        progress = tqdm(
            total=len(variants_idxs),
            desc="aligning log with decomposition/recomposition, completed variants :: ",
        )

    # Process variants one by one
    variants_to_process = []

    for index_variant, variant in enumerate(variants_idxs):
        variants_to_process.append((variant, log[variants_idxs[variant][0]], index_variant))

    all_alignments = [None] * len(variants_to_process)  # Pre-allocate result list

    thm = thread_utils.Pm4pyThreadManager()
    def _compute(variant_info, nets, results):
        idx = variant_info[2]
        results[idx] = apply_trace(variant_info[1], nets, parameters=parameters)
        if progress is not None:
            progress.update()

    for variant_info in variants_to_process:
        thm.submit(_compute, variant_info, list_nets, all_alignments)

    thm.join()

    # Map alignments back to original traces
    al_idx = {}
    for index_variant, variant in enumerate(variants_idxs):
        for trace_idx in variants_idxs[variant]:
            al_idx[trace_idx] = all_alignments[index_variant]

    alignments = [al_idx[i] for i in range(len(log))]

    # Close progress bar
    if progress is not None:
        progress.close()
        del progress

    return alignments


def get_acache(cons_nets):
    """
    Calculates the A-Cache of the given decomposition

    Parameters
    --------------
    cons_nets
        List of considered nets

    Returns
    --------------
    acache
        A-Cache
    """
    # Optimized version that pre-allocates dictionary and avoids repeated lookups
    ret = {}
    for index, el in enumerate(cons_nets):
        for lab in el[0].lvis_labels:
            if lab not in ret:
                ret[lab] = []
            ret[lab].append(index)

    return ret


def get_alres(al):
    """
    Gets a description of the alignment for the border agreement

    Parameters
    --------------
    al
        Alignment

    Returns
    --------------
    alres
        Description of the alignment
    """
    if al is not None:
        # Use a more efficient approach
        ret = {}
        for move in al["alignment"]:
            model_move = move[1][0]
            if model_move is not None and model_move != ">>":
                if model_move not in ret:
                    ret[model_move] = []

                log_move = move[1][1]
                ret[model_move].append(0 if log_move is not None and log_move != ">>" else 1)

        return ret
    return None


def order_nodes_second_round(to_visit, G0):
    """
    Orders the second round of nodes to visit to reconstruct the alignment

    Optimized version with improved algorithm
    """
    edges_cache = {}
    # Pre-calculate all edges for faster lookups
    for i in range(len(to_visit)):
        for j in range(i + 1, len(to_visit)):
            node_i, node_j = to_visit[i], to_visit[j]
            if node_i != node_j:
                # Cache edge checks to avoid repeated graph lookups
                key_ij = (node_j, node_i)
                key_ji = (node_i, node_j)

                if key_ij not in edges_cache:
                    edges_cache[key_ij] = any(e[0] == node_j and e[1] == node_i for e in G0.edges)

                if key_ji not in edges_cache:
                    edges_cache[key_ji] = any(e[0] == node_i and e[1] == node_j for e in G0.edges)

    # Now use the cache for a more efficient sorting algorithm
    swapped = True
    while swapped:
        swapped = False
        for i in range(len(to_visit) - 1):
            node_i, node_j = to_visit[i], to_visit[i + 1]
            if node_i != node_j:
                key_ij = (node_j, node_i)
                key_ji = (node_i, node_j)

                if edges_cache.get(key_ij, False) and not edges_cache.get(key_ji, False):
                    to_visit[i], to_visit[i + 1] = to_visit[i + 1], to_visit[i]
                    swapped = True

    return to_visit


def recompose_alignment(cons_nets, cons_nets_result):
    """
    Alignment recomposition

    Optimized version with more efficient graph operations
    """
    # Create graph of valid nodes
    G0 = nx_utils.DiGraph()
    valid_nodes = [i for i in range(len(cons_nets_result)) if cons_nets_result[i] is not None]

    # Add nodes in one batch
    G0.add_nodes_from(valid_nodes)

    # Efficiently add edges
    edges_to_add = []
    for i in valid_nodes:
        for j in valid_nodes:
            if i != j and cons_nets_result[i]["alignment"][-1][1] == cons_nets_result[j]["alignment"][0][1]:
                edges_to_add.append((i, j))

    G0.add_edges_from(edges_to_add)

    # Find starting nodes (nodes with initial markings)
    to_visit = [i for i in range(len(cons_nets)) if len(list(cons_nets[i][1])) > 0]
    visited = set()
    overall_ali = []
    count = 0

    # Process the first round of nodes
    while to_visit:
        curr = to_visit.pop(0)
        output_edges = [e for e in G0.edges if e[0] == curr]
        to_visit.extend(e[1] for e in output_edges)

        sind = 1 if count > 0 else 0

        if cons_nets_result[curr] is not None:
            overall_ali.extend(cons_nets_result[curr]["alignment"][sind:])

        visited.add(curr)
        count += 1

    # Process remaining nodes
    all_available = [i for i in range(len(cons_nets_result)) if cons_nets_result[i] is not None]
    to_visit = [x for x in all_available if x not in visited]
    to_visit = order_nodes_second_round(to_visit, G0)

    added = set()
    while to_visit:
        curr = to_visit.pop(0)
        if curr not in visited:
            output_edges = [e for e in G0.edges if e[0] == curr]
            to_visit.extend(e[1] for e in output_edges)

            sind = 1 if count > 0 else 0

            if cons_nets_result[curr] is not None:
                for y in cons_nets_result[curr]["alignment"][sind:]:
                    if y not in added:
                        overall_ali.append(y)
                        added.add(y)

            visited.add(curr)
        count += 1

    return overall_ali


def apply_trace(trace, list_nets, parameters=None):
    """
    Align a trace against a decomposition

    Optimized version with improved algorithms
    """
    if parameters is None:
        parameters = {}

    max_align_time_trace = exec_utils.get_param_value(
        Parameters.PARAM_MAX_ALIGN_TIME_TRACE, parameters, sys.maxsize
    )
    threshold_border_agreement = exec_utils.get_param_value(
        Parameters.PARAM_THRESHOLD_BORDER_AGREEMENT, parameters, 100000000
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    icache = exec_utils.get_param_value(Parameters.ICACHE, parameters, dict())
    mcache = exec_utils.get_param_value(Parameters.MCACHE, parameters, dict())

    # Make a shallow copy to avoid modifying the original
    cons_nets = list(list_nets)
    acache = get_acache(cons_nets)

    # Pre-allocate arrays
    cons_nets_result = []
    cons_nets_alres = []
    cons_nets_costs = []
    max_val_alres = 0
    start_time = time.time()

    # Extract activities from trace once to avoid repeated operations
    trace_activities = {x[activity_key] for x in trace}

    i = 0
    while i < len(cons_nets):
        this_time = time.time()
        if this_time - start_time > max_align_time_trace:
            # Time limit exceeded
            return None

        net, im, fm = cons_nets[i]

        # Find intersection of trace activities and net labels more efficiently
        relevant_activities = trace_activities.intersection(net.lvis_labels)

        # Create projection more efficiently
        proj = Trace([x for x in trace if x[activity_key] in relevant_activities])

        if proj:
            # Use tuple for immutable key
            acti = tuple(x[activity_key] for x in proj)
            tup = (cons_nets[i], acti)

            if tup not in icache:
                al, cf = align(proj, net, im, fm, parameters=parameters)
                alres = get_alres(al)
                icache[tup] = (al, cf, alres)

            al, cf, alres = icache[tup]
            cons_nets_result.append(al)
            cons_nets_alres.append(alres)
            cons_nets_costs.append(cf)

            if this_time - start_time > max_align_time_trace:
                return None

            # Compute max_val_alres more efficiently
            if alres:
                current_max = max(max(y) for y in alres.values() if y)
                max_val_alres = max(max_val_alres, current_max)

            border_disagreements = 0
            if max_val_alres > 0:
                comp_to_merge = set()

                # Only process relevant activities
                for act in relevant_activities:
                    for ind in acache.get(act, []):
                        if ind >= i:
                            break

                        if (not cons_nets_alres[ind] or not cons_nets_alres[i] or
                                act not in cons_nets_alres[ind] or act not in cons_nets_alres[i] or
                                cons_nets_alres[ind][act] != cons_nets_alres[i][act]):
                            # Add all components to merge
                            comp_to_merge.update(acache.get(act, []))

                if comp_to_merge:
                    comp_to_merge = sorted(list(comp_to_merge), reverse=True)
                    border_disagreements += len(comp_to_merge)

                    # Check threshold early
                    if border_disagreements > threshold_border_agreement:
                        return None

                    # Use frozenset for immutable dictionary key
                    comp_to_merge_ids = frozenset(cons_nets[j][0].t_tuple for j in comp_to_merge)

                    if comp_to_merge_ids not in mcache:
                        mcache[comp_to_merge_ids] = decomp_utils.merge_sublist_nets(
                            [cons_nets[zz] for zz in comp_to_merge]
                        )

                    new_comp = mcache[comp_to_merge_ids]
                    cons_nets.append(new_comp)

                    # Remove components more efficiently
                    for z in sorted(comp_to_merge, reverse=True):
                        if z < i:
                            i -= 1
                        if z <= i:
                            cons_nets_result.pop(z)
                            cons_nets_alres.pop(z)
                            cons_nets_costs.pop(z)
                        cons_nets.pop(z)

                    # Recalculate the activity cache
                    acache = get_acache(cons_nets)
                    continue
        else:
            cons_nets_result.append(None)
            cons_nets_alres.append(None)
            cons_nets_costs.append(None)

        i += 1

    if time.time() - start_time > max_align_time_trace:
        return None

    # Recompose the alignment
    alignment = recompose_alignment(cons_nets, cons_nets_result)

    # Build cost dictionary more efficiently
    overall_cost_dict = {}
    for cf in filter(None, cons_nets_costs):
        overall_cost_dict.update(cf)

    # Calculate total cost
    cost = sum(overall_cost_dict[el] for el in alignment)

    # Extract alignment
    alignment = [x[1] for x in alignment]

    if time.time() - start_time > max_align_time_trace:
        return None

    # Build result with fitness if needed
    res = {"cost": cost, "alignment": alignment}

    best_worst_cost = exec_utils.get_param_value(
        Parameters.BEST_WORST_COST, parameters, None
    )

    if best_worst_cost is not None and trace:
        cost1 = cost // utils.STD_MODEL_LOG_MOVE_COST
        fitness = 1.0 - cost1 / (best_worst_cost + len(trace))
        res["fitness"] = fitness
        res["bwc"] = (best_worst_cost + len(trace)) * utils.STD_MODEL_LOG_MOVE_COST

    return res


def align(trace, petri_net, initial_marking, final_marking, parameters=None):
    """
    Perform alignment using state_equation_a_star
    """
    if parameters is None:
        parameters = {}

    new_parameters = copy(parameters)
    new_parameters[
        state_equation_a_star.Parameters.RETURN_SYNC_COST_FUNCTION
    ] = True
    new_parameters[
        state_equation_a_star.Parameters.PARAM_ALIGNMENT_RESULT_IS_SYNC_PROD_AWARE
    ] = True

    aligned_trace, cost_function = state_equation_a_star.apply(
        trace,
        petri_net,
        initial_marking,
        final_marking,
        parameters=new_parameters,
    )

    # Create cost function dictionary more efficiently
    cf = {((x.name[0], x.name[1]), (x.label[0], x.label[1])): cost_function[x]
          for x in cost_function}

    return aligned_trace, cf



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
from enum import Enum
from typing import Any, Optional, Dict, Union

import polars as pl

from pm4py.util import constants, xes_constants, exec_utils


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    MIN_REP = "min_rep"
    MAX_REP = "max_rep"


def apply(
    df: pl.LazyFrame,
    value: Any,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filters the trace of the LazyFrame where the given attribute value is repeated
    (in a range of repetitions that is specified by the user)

    Parameters
    ----------------
    df
        LazyFrame
    value
        Value that is investigated
    parameters
        Parameters of the filter, including:
        - Parameters.ATTRIBUTE_KEY => the attribute key
        - Parameters.MIN_REP => minimum number of repetitions
        - Parameters.MAX_REP => maximum number of repetitions
        - Parameters.CASE_ID_KEY => the columns of the LazyFrame that is the case identifier

    Returns
    ----------------
    filtered_df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    attribute_key = exec_utils.get_param_value(
        Parameters.ATTRIBUTE_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    min_rep = exec_utils.get_param_value(Parameters.MIN_REP, parameters, 2)
    max_rep = exec_utils.get_param_value(
        Parameters.MAX_REP, parameters, sys.maxsize
    )

    # Filter events with the specified value
    filtered_df = df.filter(pl.col(attribute_key) == value)
    
    # Count occurrences per case
    case_counts = (
        filtered_df
        .group_by(case_id_key)
        .count()
        .filter(
            (pl.col("count") >= min_rep) & (pl.col("count") <= max_rep)
        )
        .select(case_id_key)
    )

    # Keep only cases that meet the repetition criteria
    ret = df.join(case_counts, on=case_id_key, how="inner")

    return ret


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Dict, Optional, Any, Union

import pandas as pd

from pm4py.statistics.traces.cycle_time.util import compute
from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


def apply(
    df: pd.DataFrame,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Computes the cycle time starting from a Pandas dataframe

    The definition that has been followed is the one proposed in:
    https://www.presentationeze.com/presentations/lean-manufacturing-just-in-time/lean-manufacturing-just-in-time-full-details/process-cycle-time-analysis/calculate-cycle-time/#:~:text=Cycle%20time%20%3D%20Average%20time%20between,is%2024%20minutes%20on%20average.

    So:
    Cycle time  = Average time between completion of units.

    Example taken from the website:
    Consider a manufacturing facility, which is producing 100 units of product per 40 hour week.
    The average throughput rate is 1 unit per 0.4 hours, which is one unit every 24 minutes.
    Therefore the cycle time is 24 minutes on average.

    Parameters
    ------------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.START_TIMESTAMP_KEY => the attribute acting as start timestamp
        - Parameters.TIMESTAMP_KEY => the attribute acting as timestamp
        - Parameters.CASE_ID_KEY => the attribute acting as case identifier

    Returns
    ------------------
    cycle_time
        Cycle time
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY,
        parameters,
        xes_constants.DEFAULT_TIMESTAMP_KEY,
    )
    start_timestamp_key = exec_utils.get_param_value(
        Parameters.START_TIMESTAMP_KEY, parameters, timestamp_key
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    events = [
        (x[start_timestamp_key].timestamp(), x[timestamp_key].timestamp())
        for x in df[list({start_timestamp_key, timestamp_key})].to_dict(
            "records"
        )
    ]

    return compute.cycle_time(events, df[case_id_key].nunique())



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Adds for each object the number of events in its lifecycle as feature

    Parameters
    -----------------
    ocel
        OCEL
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    data
        Values of the added features
    feature_names
        Names of the added features
    """
    if parameters is None:
        parameters = {}

    ordered_objects = (
        parameters["ordered_objects"]
        if "ordered_objects" in parameters
        else ocel.objects[ocel.object_id_column].to_numpy()
    )

    lifecycle_length = (
        ocel.relations.groupby(ocel.object_id_column).size().to_dict()
    )

    data = []
    feature_names = ["@@object_lifecycle_length"]

    for obj in ordered_objects:
        if obj in lifecycle_length:
            data.append([float(lifecycle_length[obj])])
        else:
            data.append([0.0])

    return data, feature_names



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import time
import uuid
from itertools import product

import pm4py.objects.conversion.process_tree.variants.to_petri_net as pt_to_pn
from pm4py.objects.petri_net.obj import Marking
from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils.petri_utils import (
    add_arc_from_to,
    remove_place,
)
from pm4py.objects.powl.obj import (
    Transition,
    SilentTransition,
    StrictPartialOrder,
    OperatorPOWL,
    FrequentTransition,
)
from pm4py.objects.petri_net.utils import reduction
from pm4py.objects.process_tree.obj import Operator


def recursively_add_tree(
    powl,
    net,
    initial_entity_subtree,
    final_entity_subtree,
    counts,
    rec_depth,
    force_add_skip=False,
):
    """
    Recursively add the submodels to the Petri net

    Parameters
    -----------
    powl
        POWL model
    net
        Petri net
    initial_entity_subtree
        Initial entity (place/transition) that should be attached from the subtree
    final_entity_subtree
        Final entity (place/transition) that should be attached from the subtree
    counts
        Counts object (keeps the number of places, transitions and hidden transitions)
    rec_depth
        Recursion depth of the current iteration
    force_add_skip
        Boolean value that tells if the addition of a skip is mandatory

    Returns
    ----------
    net
        Updated Petri net
    counts
        Updated counts object (keeps the number of places, transitions and hidden transitions)
    final_place
        Last place added in this recursion
    """
    if type(initial_entity_subtree) is PetriNet.Transition:
        initial_place = get_new_place(counts)
        net.places.add(initial_place)
        add_arc_from_to(initial_entity_subtree, initial_place, net)
    else:
        initial_place = initial_entity_subtree
    if (
        final_entity_subtree is not None
        and type(final_entity_subtree) is PetriNet.Place
    ):
        final_place = final_entity_subtree
    else:
        final_place = get_new_place(counts)
        net.places.add(final_place)
        if (
            final_entity_subtree is not None
            and type(final_entity_subtree) is PetriNet.Transition
        ):
            add_arc_from_to(final_place, final_entity_subtree, net)

    if force_add_skip:
        invisible = get_new_hidden_trans(counts, type_trans="skip")
        add_arc_from_to(initial_place, invisible, net)
        add_arc_from_to(invisible, final_place, net)

    if isinstance(powl, Transition):
        if isinstance(powl, SilentTransition):
            petri_trans = get_new_hidden_trans(counts, type_trans="skip")
        elif isinstance(powl, FrequentTransition):
            petri_trans = get_transition(
                counts,
                powl.label,
                powl.activity,
                powl.skippable,
                powl.selfloop,
            )
        else:
            petri_trans = get_transition(counts, powl.label, powl.label)
        net.transitions.add(petri_trans)
        add_arc_from_to(initial_place, petri_trans, net)
        add_arc_from_to(petri_trans, final_place, net)

    elif isinstance(powl, OperatorPOWL):
        tree_children = powl.children
        if powl.operator == Operator.XOR:
            for subtree in tree_children:
                net, counts, intermediate_place = recursively_add_tree(
                    subtree,
                    net,
                    initial_place,
                    final_place,
                    counts,
                    rec_depth + 1,
                )
        elif powl.operator == Operator.LOOP:
            new_initial_place = get_new_place(counts)
            net.places.add(new_initial_place)
            init_loop_trans = get_new_hidden_trans(
                counts, type_trans="init_loop"
            )
            net.transitions.add(init_loop_trans)
            add_arc_from_to(initial_place, init_loop_trans, net)
            add_arc_from_to(init_loop_trans, new_initial_place, net)
            initial_place = new_initial_place
            loop_trans = get_new_hidden_trans(counts, type_trans="loop")
            net.transitions.add(loop_trans)

            exit_node = SilentTransition()
            do = tree_children[0]
            redo = tree_children[1]

            net, counts, int1 = recursively_add_tree(
                do, net, initial_place, None, counts, rec_depth + 1
            )
            net, counts, int2 = recursively_add_tree(
                redo, net, int1, None, counts, rec_depth + 1
            )
            net, counts, int3 = recursively_add_tree(
                exit_node, net, int1, final_place, counts, rec_depth + 1
            )

            looping_place = int2

            add_arc_from_to(looping_place, loop_trans, net)
            add_arc_from_to(loop_trans, initial_place, net)

    elif isinstance(powl, StrictPartialOrder):
        transitive_reduction = powl.order.get_transitive_reduction()
        tree_children = list(powl.children)
        tau_split = get_new_hidden_trans(counts, type_trans="tauSplit")
        net.transitions.add(tau_split)
        add_arc_from_to(initial_place, tau_split, net)
        tau_join = get_new_hidden_trans(counts, type_trans="tauJoin")
        net.transitions.add(tau_join)
        add_arc_from_to(tau_join, final_place, net)

        init_trans = []
        final_trans = []
        start_nodes = transitive_reduction.get_start_nodes()
        end_nodes = transitive_reduction.get_end_nodes()
        for subtree in tree_children:
            i_trans = get_new_hidden_trans(counts, type_trans="init_par")
            net.transitions.add(i_trans)
            if subtree in start_nodes:
                i_place = get_new_place(counts)
                net.places.add(i_place)
                add_arc_from_to(tau_split, i_place, net)

                add_arc_from_to(i_place, i_trans, net)

            f_trans = get_new_hidden_trans(counts, type_trans="final_par")
            net.transitions.add(f_trans)
            if subtree in end_nodes:
                f_place = get_new_place(counts)
                net.places.add(f_place)
                add_arc_from_to(f_trans, f_place, net)
                add_arc_from_to(f_place, tau_join, net)

            net, counts, intermediate_place = recursively_add_tree(
                subtree, net, i_trans, f_trans, counts, rec_depth + 1
            )
            init_trans.append(i_trans)
            final_trans.append(f_trans)

        n = range(len(tree_children))
        for i, j in product(n, n):
            if transitive_reduction.is_edge_id(i, j):
                new_place = get_new_place(counts)
                net.places.add(new_place)
                add_arc_from_to(final_trans[i], new_place, net)
                add_arc_from_to(new_place, init_trans[j], net)

    return net, counts, final_place


def apply(powl, parameters=None):
    """
    Apply from POWL model to Petri net

    Parameters
    -----------
    powl
        POWL model
    parameters
        Parameters of the algorithm

    Returns
    -----------
    net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """

    counts = pt_to_pn.Counts()
    net = PetriNet("imdf_net_" + str(time.time()))
    initial_marking = Marking()
    final_marking = Marking()
    source = get_new_place(counts)
    source.name = "source"
    sink = get_new_place(counts)
    sink.name = "sink"
    net.places.add(source)
    net.places.add(sink)
    initial_marking[source] = 1
    final_marking[sink] = 1
    initial_mandatory = True  # check_tau_mandatory_at_initial_marking(powl)
    final_mandatory = True  # check_tau_mandatory_at_final_marking(powl)
    if initial_mandatory:
        initial_place = get_new_place(counts)
        net.places.add(initial_place)
        tau_initial = get_new_hidden_trans(counts, type_trans="tau")
        net.transitions.add(tau_initial)
        add_arc_from_to(source, tau_initial, net)
        add_arc_from_to(tau_initial, initial_place, net)
    else:
        initial_place = source
    if final_mandatory:
        final_place = get_new_place(counts)
        net.places.add(final_place)
        tau_final = get_new_hidden_trans(counts, type_trans="tau")
        net.transitions.add(tau_final)
        add_arc_from_to(final_place, tau_final, net)
        add_arc_from_to(tau_final, sink, net)
    else:
        final_place = sink

    net, counts, last_added_place = recursively_add_tree(
        powl, net, initial_place, final_place, counts, 0
    )

    reduction.apply_simple_reduction(net)

    places = list(net.places)
    for place in places:
        if len(place.out_arcs) == 0 and place not in final_marking:
            remove_place(net, place)
        if len(place.in_arcs) == 0 and place not in initial_marking:
            remove_place(net, place)

    return net, initial_marking, final_marking


def get_new_place(counts):
    """
    Create a new place in the Petri net
    """
    counts.inc_places()
    return PetriNet.Place("p_" + str(counts.num_places))


def get_new_hidden_trans(counts, type_trans="unknown"):
    """
    Create a new hidden transition in the Petri net
    """
    counts.inc_no_hidden()
    return PetriNet.Transition(type_trans + "_" + str(counts.num_hidden), None)


def get_transition(counts, label, activity, skippable=False, selfloop=False):
    """
    Create a transitions with the specified label in the Petri net
    """
    counts.inc_no_visible()
    return PetriNet.Transition(
        str(uuid.uuid4()),
        label,
        properties={
            "activity": activity,
            "skippable": skippable,
            "selfloop": selfloop,
        },
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from copy import copy

from pm4py.algo.conformance.alignments.petri_net import variants
from pm4py.objects.petri_net.utils import align_utils, check_soundness
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.util.xes_constants import DEFAULT_NAME_KEY, DEFAULT_TRACEID_KEY
from pm4py.objects.log.obj import Trace, Event
import time
from pm4py.util.lp import solver
from pm4py.util import exec_utils, thread_utils
from enum import Enum
import sys
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    PARAMETER_CONSTANT_CASEID_KEY,
    CASE_CONCEPT_NAME,
)
import importlib.util
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream, Trace
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.util import typing, constants, pandas_utils
import pandas as pd


class Variants(Enum):
    VERSION_STATE_EQUATION_A_STAR = variants.state_equation_a_star
    VERSION_DIJKSTRA_NO_HEURISTICS = variants.dijkstra_no_heuristics
    VERSION_DIJKSTRA_LESS_MEMORY = variants.dijkstra_less_memory
    VERSION_DISCOUNTED_A_STAR = variants.discounted_a_star

class Parameters(Enum):
    PARAM_TRACE_COST_FUNCTION = "trace_cost_function"
    PARAM_MODEL_COST_FUNCTION = "model_cost_function"
    PARAM_SYNC_COST_FUNCTION = "sync_cost_function"
    PARAM_ALIGNMENT_RESULT_IS_SYNC_PROD_AWARE = "ret_tuple_as_trans_desc"
    PARAM_TRACE_NET_COSTS = "trace_net_costs"
    TRACE_NET_CONSTR_FUNCTION = "trace_net_constr_function"
    TRACE_NET_COST_AWARE_CONSTR_FUNCTION = (
        "trace_net_cost_aware_constr_function"
    )
    PARAM_MAX_ALIGN_TIME_TRACE = "max_align_time_trace"
    PARAM_MAX_ALIGN_TIME = "max_align_time"
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    VARIANTS_IDX = "variants_idx"
    SHOW_PROGRESS_BAR = "show_progress_bar"
    CORES = "cores"
    BEST_WORST_COST_INTERNAL = "best_worst_cost_internal"
    FITNESS_ROUND_DIGITS = "fitness_round_digits"
    SYNCHRONOUS = "synchronous_dijkstra"
    EXPONENT="theta"
    ENABLE_BEST_WORST_COST = "enable_best_worst_cost"


def __variant_mapper(variant):
    if type(variant) is str:
        if variant == "Variants.VERSION_STATE_EQUATION_A_STAR":
            variant = Variants.VERSION_STATE_EQUATION_A_STAR
        elif variant == "Variants.VERSION_DIJKSTRA_NO_HEURISTICS":
            variant = Variants.VERSION_DIJKSTRA_NO_HEURISTICS
        elif variant == "Variants.VERSION_DIJKSTRA_LESS_MEMORY":
            variant = Variants.VERSION_DIJKSTRA_LESS_MEMORY

    return variant


DEFAULT_VARIANT = Variants.VERSION_DIJKSTRA_LESS_MEMORY
if solver.DEFAULT_LP_SOLVER_VARIANT is not None:
    DEFAULT_VARIANT = __variant_mapper(constants.DEFAULT_ALIGNMENTS_VARIANT)

VERSION_STATE_EQUATION_A_STAR = Variants.VERSION_STATE_EQUATION_A_STAR
VERSION_DIJKSTRA_NO_HEURISTICS = Variants.VERSION_DIJKSTRA_NO_HEURISTICS
VERSION_DIJKSTRA_LESS_MEMORY = Variants.VERSION_DIJKSTRA_LESS_MEMORY

VERSIONS = {
    Variants.VERSION_DIJKSTRA_NO_HEURISTICS,
    Variants.VERSION_DIJKSTRA_NO_HEURISTICS,
    Variants.VERSION_DIJKSTRA_LESS_MEMORY,
}


def apply(
    obj: Union[EventLog, EventStream, pd.DataFrame, Trace],
    petri_net: PetriNet,
    initial_marking: Marking,
    final_marking: Marking,
    parameters: Optional[Dict[Any, Any]] = None,
    variant=DEFAULT_VARIANT,
) -> Union[typing.AlignmentResult, typing.ListAlignments]:
    if parameters is None:
        parameters = {}
    if isinstance(obj, Trace):
        return apply_trace(
            obj,
            petri_net,
            initial_marking,
            final_marking,
            parameters=parameters,
            variant=variant,
        )
    else:
        return apply_log(
            obj,
            petri_net,
            initial_marking,
            final_marking,
            parameters=parameters,
            variant=variant,
        )


def apply_trace(
    trace,
    petri_net,
    initial_marking,
    final_marking,
    parameters=None,
    variant=DEFAULT_VARIANT,
):
    """
    apply alignments to a trace
    Parameters
    -----------
    trace
        :class:`pm4py.log.log.Trace` trace of events
    petri_net
        :class:`pm4py.objects.petri.petrinet.PetriNet` the model to use for the alignment
    initial_marking
        :class:`pm4py.objects.petri.petrinet.Marking` initial marking of the net
    final_marking
        :class:`pm4py.objects.petri.petrinet.Marking` final marking of the net
    variant
        selected variant of the algorithm, possible values: {\'Variants.VERSION_STATE_EQUATION_A_STAR, Variants.VERSION_DIJKSTRA_NO_HEURISTICS \'}
    parameters
        :class:`dict` parameters of the algorithm, for key \'state_equation_a_star\':
            Parameters.ACTIVITY_KEY -> Attribute in the log that contains the activity
            Parameters.PARAM_MODEL_COST_FUNCTION ->
            mapping of each transition in the model to corresponding synchronous costs
            Parameters.PARAM_SYNC_COST_FUNCTION ->
            mapping of each transition in the model to corresponding model cost
            Parameters.PARAM_TRACE_COST_FUNCTION ->
            mapping of each index of the trace to a positive cost value
    Returns
    -----------
    alignment
        :class:`dict` with keys **alignment**, **cost**, **visited_states**, **queued_states** and
        **traversed_arcs**
        The alignment is a sequence of labels of the form (a,t), (a,>>), or (>>,t)
        representing synchronous/log/model-moves.
    """
    if parameters is None:
        parameters = copy({PARAMETER_CONSTANT_ACTIVITY_KEY: DEFAULT_NAME_KEY})

    variant = __variant_mapper(variant)
    parameters = copy(parameters)

    enable_best_worst_cost = exec_utils.get_param_value(
        Parameters.ENABLE_BEST_WORST_COST, parameters, True
    )

    ali = exec_utils.get_variant(variant).apply(
        trace, petri_net, initial_marking, final_marking, parameters=parameters
    )

    trace_cost_function = exec_utils.get_param_value(
        Parameters.PARAM_TRACE_COST_FUNCTION, parameters, []
    )
    # Instead of using the length of the trace, use the sum of the trace cost
    # function
    trace_cost_function_sum = sum(trace_cost_function)

    if enable_best_worst_cost:
        best_worst_cost = exec_utils.get_param_value(
            Parameters.BEST_WORST_COST_INTERNAL,
            parameters,
            __get_best_worst_cost(
                petri_net, initial_marking, final_marking, variant, parameters
            ),
        )

        if ali is not None and best_worst_cost is not None:
            ltrace_bwc = trace_cost_function_sum + best_worst_cost

            fitness_num = ali["cost"] // align_utils.STD_MODEL_LOG_MOVE_COST
            fitness_den = ltrace_bwc // align_utils.STD_MODEL_LOG_MOVE_COST
            fitness = 1 - fitness_num / fitness_den if fitness_den > 0 else 0

            ali["fitness"] = fitness
            # returning also the best worst cost, for log fitness computation
            ali["bwc"] = ltrace_bwc

    return ali


def apply_log(
    log,
    petri_net,
    initial_marking,
    final_marking,
    parameters=None,
    variant=DEFAULT_VARIANT,
):
    """
    apply alignments to a log
    Parameters
    -----------
    log
        object of the form :class:`pm4py.log.log.EventLog` event log
    petri_net
        :class:`pm4py.objects.petri.petrinet.PetriNet` the model to use for the alignment
    initial_marking
        :class:`pm4py.objects.petri.petrinet.Marking` initial marking of the net
    final_marking
        :class:`pm4py.objects.petri.petrinet.Marking` final marking of the net
    variant
        selected variant of the algorithm, possible values: {\'Variants.VERSION_STATE_EQUATION_A_STAR, Variants.VERSION_DIJKSTRA_NO_HEURISTICS \'}
    parameters
        :class:`dict` parameters of the algorithm,

    Returns
    -----------
    alignment
        :class:`list` of :class:`dict` with keys **alignment**, **cost**, **visited_states**, **queued_states** and
        **traversed_arcs**
        The alignment is a sequence of labels of the form (a,t), (a,>>), or (>>,t)
        representing synchronous/log/model-moves.
    """
    if parameters is None:
        parameters = dict()

    if solver.DEFAULT_LP_SOLVER_VARIANT is not None:
        if not check_soundness.check_easy_soundness_net_in_fin_marking(
            petri_net, initial_marking, final_marking
        ):
            raise Exception(
                "trying to apply alignments on a Petri net that is not a easy sound net!!"
            )

    enable_best_worst_cost = exec_utils.get_param_value(
        Parameters.ENABLE_BEST_WORST_COST, parameters, True
    )

    variant = __variant_mapper(variant)

    start_time = time.time()
    max_align_time = exec_utils.get_param_value(
        Parameters.PARAM_MAX_ALIGN_TIME, parameters, sys.maxsize
    )
    max_align_time_case = exec_utils.get_param_value(
        Parameters.PARAM_MAX_ALIGN_TIME_TRACE, parameters, sys.maxsize
    )

    variants_idxs, one_tr_per_var = __get_variants_structure(log, parameters)
    progress = __get_progress_bar(len(one_tr_per_var), parameters)

    if enable_best_worst_cost:
        best_worst_cost = __get_best_worst_cost(
            petri_net, initial_marking, final_marking, variant, parameters
        )
        parameters[Parameters.BEST_WORST_COST_INTERNAL] = best_worst_cost

    is_dijkstra = str(variant) in {"Variants.VERSION_DIJKSTRA_LESS_MEMORY", "Variants.VERSION_DIJKSTRA_NO_HEURISTICS"}
    thm = thread_utils.Pm4pyThreadManager()

    all_alignments = [None] * len(one_tr_per_var)

    def _compute(idx, trace, params, results):
        results[idx] = apply_trace(
            trace,
            petri_net,
            initial_marking,
            final_marking,
            parameters=params,
            variant=variant,
        )
        if progress is not None:
            progress.update()

    for idx, trace in enumerate(one_tr_per_var):
        this_max_align_time = min(
            max_align_time_case,
            (max_align_time - (time.time() - start_time)) * 0.5,
        )
        parameters[Parameters.PARAM_MAX_ALIGN_TIME_TRACE] = this_max_align_time
        thm.submit(
            _compute,
            idx,
            trace,
            copy(parameters),
            all_alignments,
        )

    thm.join()

    alignments = __form_alignments(variants_idxs, all_alignments)
    __close_progress_bar(progress)

    return alignments


def apply_multiprocessing(
    log,
    petri_net,
    initial_marking,
    final_marking,
    parameters=None,
    variant=DEFAULT_VARIANT,
):
    """
    Applies the alignments using a process pool (multiprocessing)

    Parameters
    ---------------
    log
        Event log
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    parameters
        Parameters of the algorithm

    Returns
    ----------------
    aligned_traces
        Alignments
    """
    if parameters is None:
        parameters = {}

    import multiprocessing

    variant = __variant_mapper(variant)

    num_cores = exec_utils.get_param_value(
        Parameters.CORES, parameters, multiprocessing.cpu_count() - 2
    )

    enable_best_worst_cost = exec_utils.get_param_value(
        Parameters.ENABLE_BEST_WORST_COST, parameters, True
    )

    variants_idxs, one_tr_per_var = __get_variants_structure(log, parameters)

    if enable_best_worst_cost:
        best_worst_cost = __get_best_worst_cost(
            petri_net, initial_marking, final_marking, variant, parameters
        )
        parameters[Parameters.BEST_WORST_COST_INTERNAL] = best_worst_cost

    all_alignments = []

    from concurrent.futures import ProcessPoolExecutor

    with ProcessPoolExecutor(max_workers=num_cores) as executor:
        futures = []
        for trace in one_tr_per_var:
            futures.append(
                executor.submit(
                    apply_trace,
                    trace,
                    petri_net,
                    initial_marking,
                    final_marking,
                    parameters,
                    str(variant),
                )
            )
        progress = __get_progress_bar(len(one_tr_per_var), parameters)
        if progress is not None:
            alignments_ready = 0
            while alignments_ready != len(futures):
                current = 0
                for index, variant in enumerate(futures):
                    current = current + 1 if futures[index].done() else current
                if current > alignments_ready:
                    for i in range(0, current - alignments_ready):
                        progress.update()
                alignments_ready = current
        for index, variant in enumerate(futures):
            all_alignments.append(futures[index].result())
        __close_progress_bar(progress)

    alignments = __form_alignments(variants_idxs, all_alignments)

    return alignments


def __get_best_worst_cost(
    petri_net, initial_marking, final_marking, variant, parameters
):
    parameters_best_worst = copy(parameters)

    best_worst_cost = exec_utils.get_variant(variant).get_best_worst_cost(
        petri_net,
        initial_marking,
        final_marking,
        parameters=parameters_best_worst,
    )

    return best_worst_cost


def __get_variants_structure(log, parameters):
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )

    variants_idxs = {}
    one_tr_per_var = []

    if pandas_utils.check_is_pandas_dataframe(log):
        case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
        )
        traces = pandas_utils.get_traces(log, case_id_key, activity_key)
        for idx, trace in enumerate(traces):
            if trace not in variants_idxs:
                variants_idxs[trace] = [idx]
                case = Trace()
                for act in trace:
                    case.append(Event({activity_key: act}))
                one_tr_per_var.append(case)
            else:
                variants_idxs[trace].append(idx)
    else:
        log = log_converter.apply(
            log,
            variant=log_converter.Variants.TO_EVENT_LOG,
            parameters=parameters,
        )
        for idx, case in enumerate(log):
            trace = tuple(x[activity_key] for x in case)
            if trace not in variants_idxs:
                variants_idxs[trace] = [idx]
                one_tr_per_var.append(case)
            else:
                variants_idxs[trace].append(idx)

    return variants_idxs, one_tr_per_var


def __get_progress_bar(num_variants, parameters):
    show_progress_bar = exec_utils.get_param_value(
        Parameters.SHOW_PROGRESS_BAR, parameters, constants.SHOW_PROGRESS_BAR
    )
    progress = None
    if (
        importlib.util.find_spec("tqdm")
        and show_progress_bar
        and num_variants > 1
    ):
        from tqdm.auto import tqdm

        progress = tqdm(
            total=num_variants, desc="aligning log, completed variants :: "
        )
    return progress


def __form_alignments(variants_idxs, all_alignments):
    al_idx = {}
    for index_variant, variant in enumerate(variants_idxs):
        for trace_idx in variants_idxs[variant]:
            al_idx[trace_idx] = all_alignments[index_variant]

    alignments = []
    for i in range(len(al_idx)):
        alignments.append(al_idx[i])

    return alignments


def __close_progress_bar(progress):
    if progress is not None:
        progress.close()
    del progress


def get_diagnostics_dataframe(log, align_output, parameters=None):
    """
    Gets the diagnostics results of alignments (of a log) in a dataframe

    Parameters
    --------------
    log
        Event log
    align_output
        Output of the alignments

    Returns
    --------------
    dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, DEFAULT_TRACEID_KEY
    )

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters)

    diagn_stream = []

    for index in range(len(log)):
        case_id = log[index].attributes[case_id_key]

        cost = align_output[index]["cost"]
        fitness = align_output[index]["fitness"]
        is_fit = fitness == 1.0

        diagn_stream.append(
            {
                "case_id": case_id,
                "cost": cost,
                "fitness": fitness,
                "is_fit": is_fit,
            }
        )

    return pandas_utils.instantiate_dataframe(diagn_stream)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
import json
import logging
import importlib.util
from pm4py.util import exec_utils
from enum import Enum


class Parameters(Enum):
    GRAPH_POINTS = "graph_points"
    POINT_TO_SAMPLE = "points_to_sample"


def get_kde_caseduration(duration_values, parameters=None):
    """
    Gets the estimation of KDE density for the case durations calculated on the log/dataframe

    Parameters
    --------------
    duration_values
        Values of duration
    parameters
        Possible parameters of the algorithm, including:
            graph_points -> number of points to include in the graph

    Returns
    --------------
    x
        X-axis values to represent
    y
        Y-axis values to represent
    """
    if importlib.util.find_spec("scipy"):
        from scipy.stats import gaussian_kde

        if parameters is None:
            parameters = {}

        graph_points = exec_utils.get_param_value(
            Parameters.GRAPH_POINTS, parameters, 200
        )
        duration_values = sorted(duration_values)
        
        # Check if we have enough data points for KDE
        if len(duration_values) < 2:
            # Return empty or single-point data
            if len(duration_values) == 0:
                return [[], []]
            else:
                # Single value case - return a simple representation
                single_val = duration_values[0]
                return [[single_val], [1.0]]
        
        density = gaussian_kde(duration_values)
        xs1 = list(
            np.linspace(
                min(duration_values),
                max(duration_values),
                int(graph_points / 2),
            )
        )
        xs2 = list(
            np.geomspace(
                max(min(duration_values), 0.001),
                max(duration_values),
                int(graph_points / 2),
            )
        )
        xs = sorted(xs1 + xs2)

        return [xs, list(density(xs))]
    else:
        msg = "scipy is not available. graphs cannot be built!"
        logging.error(msg)
        raise Exception(msg)


def get_kde_caseduration_json(duration_values, parameters=None):
    """
    Gets the estimation of KDE density for the case durations calculated on the log/dataframe
    (expressed as JSON)

    Parameters
    --------------
    duration_values
        Values of duration
    parameters
        Possible parameters of the algorithm, including:
            graph_points: number of points to include in the graph

    Returns
    --------------
    json
        JSON representing the graph points
    """
    x, y = get_kde_caseduration(duration_values, parameters=parameters)

    ret = []
    for i in range(len(x)):
        ret.append((x[i], y[i]))

    return json.dumps(ret)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.log.util import xes
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.algo.discovery.log_skeleton import trace_skel
from pm4py.util import xes_constants
from pm4py.util import variants_util, pandas_utils
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List, Set
from pm4py.objects.log.obj import EventLog, Trace
import pandas as pd

from enum import Enum
from pm4py.util.constants import (
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    PARAMETER_CONSTANT_CASEID_KEY,
    CASE_CONCEPT_NAME,
)


class Parameters(Enum):
    # parameter for the noise threshold
    NOISE_THRESHOLD = "noise_threshold"
    # considered constraints in conformance checking among: equivalence,
    # always_after, always_before, never_together, directly_follows,
    # activ_freq
    CONSIDERED_CONSTRAINTS = "considered_constraints"
    # default choice for conformance checking
    DEFAULT_CONSIDERED_CONSTRAINTS = [
        "equivalence",
        "always_after",
        "always_before",
        "never_together",
        "directly_follows",
        "activ_freq",
    ]
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    PARAMETER_VARIANT_DELIMITER = "variant_delimiter"


NOISE_THRESHOLD = Parameters.NOISE_THRESHOLD
CONSIDERED_CONSTRAINTS = Parameters.CONSIDERED_CONSTRAINTS
DEFAULT_CONSIDERED_CONSTRAINTS = Parameters.DEFAULT_CONSIDERED_CONSTRAINTS
ACTIVITY_KEY = Parameters.ACTIVITY_KEY
PARAMETER_VARIANT_DELIMITER = Parameters.PARAMETER_VARIANT_DELIMITER


class DiscoveryOutputs(Enum):
    EQUIVALENCE = "equivalence"
    ALWAYS_AFTER = "always_after"
    ALWAYS_BEFORE = "always_before"
    NEVER_TOGETHER = "never_together"
    DIRECTLY_FOLLOWS = "directly_follows"
    ACTIV_FREQ = "activ_freq"


class Outputs(Enum):
    DEVIATIONS = "deviations"
    NO_DEV_TOTAL = "no_dev_total"
    NO_CONSTR_TOTAL = "no_constr_total"
    DEV_FITNESS = "dev_fitness"
    IS_FIT = "is_fit"


def apply_log(
    log: Union[EventLog, pd.DataFrame],
    model: Dict[str, Any],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Set[Any]]:
    """
    Apply log-skeleton based conformance checking given an event log
    and a log-skeleton model

    Parameters
    --------------
    log
        Event log
    model
        Log-skeleton model
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY
        - Parameters.CONSIDERED_CONSTRAINTS, among: equivalence, always_after, always_before, never_together, directly_follows, activ_freq

    Returns
    --------------
    aligned_traces
        Conformance checking results for each trace:
        - Outputs.IS_FIT => boolean that tells if the trace is perfectly fit according to the model
        - Outputs.DEV_FITNESS => deviation based fitness (between 0 and 1; the more the trace is near to 1 the more fit is)
        - Outputs.DEVIATIONS => list of deviations in the model
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )

    if pandas_utils.check_is_pandas_dataframe(log):
        case_id_key = exec_utils.get_param_value(
            Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
        )
        traces = pandas_utils.get_traces(log, case_id_key, activity_key)
    else:
        traces = [tuple(y[activity_key] for y in x) for x in log]
    grouped_traces = {}
    gtk = []
    inv_idxs = {}
    for i in range(len(traces)):
        tr = traces[i]
        if tr not in grouped_traces:
            grouped_traces[tr] = []
            gtk.append(tr)
        grouped_traces[tr].append(i)
        inv_idxs[i] = gtk.index(tr)

    res0 = []
    for trace in grouped_traces:
        res0.append(apply_actlist(trace, model, parameters=parameters))

    res = []
    for i in range(len(traces)):
        res.append(res0[inv_idxs[i]])

    return res


def apply_trace(
    trace: Trace,
    model: Dict[str, Any],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> List[Set[Any]]:
    """
    Apply log-skeleton based conformance checking given a trace
    and a log-skeleton model

    Parameters
    --------------
    trace
        Trace
    model
        Log-skeleton model
    parameters
        Parameters of the algorithm, including:
        - the activity key (pm4py:param:activity_key)
        - the list of considered constraints (considered_constraints) among: equivalence, always_after, always_before, never_together, directly_follows, activ_freq

    Returns
    --------------
    aligned_trace
        Containing:
        - is_fit => boolean that tells if the trace is perfectly fit according to the model
        - dev_fitness => deviation based fitness (between 0 and 1; the more the trace is near to 1 the more fit is)
        - deviations => list of deviations in the model
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    trace = [x[activity_key] for x in trace]

    return apply_actlist(trace, model, parameters=parameters)


def apply_actlist(trace, model, parameters=None):
    """
    Apply log-skeleton based conformance checking given the list of activities of a trace
    and a log-skeleton model

    Parameters
    --------------
    trace
        List of activities of a trace
    model
        Log-skeleton model
    parameters
        Parameters of the algorithm, including:
        - the activity key (pm4py:param:activity_key)
        - the list of considered constraints (considered_constraints) among: equivalence, always_after, always_before, never_together, directly_follows, activ_freq

    Returns
    --------------
    aligned_trace
        Containing:
        - is_fit => boolean that tells if the trace is perfectly fit according to the model
        - dev_fitness => deviation based fitness (between 0 and 1; the more the trace is near to 1 the more fit is)
        - deviations => list of deviations in the model
    """
    if parameters is None:
        parameters = {}

    consid_constraints = exec_utils.get_param_value(
        Parameters.CONSIDERED_CONSTRAINTS,
        parameters,
        Parameters.DEFAULT_CONSIDERED_CONSTRAINTS.value,
    )
    trace_info = trace_skel.get_trace_info(trace)

    ret = {}
    ret[Outputs.DEVIATIONS.value] = []
    dev_total = 0
    conf_total = 0

    default_considered_constraints = (
        Parameters.DEFAULT_CONSIDERED_CONSTRAINTS.value
    )

    i = 0
    while i < len(default_considered_constraints):
        if default_considered_constraints[i] in consid_constraints:
            if (
                default_considered_constraints[i]
                == DiscoveryOutputs.ACTIV_FREQ.value
            ):
                this_constraints = {
                    x: y
                    for x, y in model[
                        default_considered_constraints[i]
                    ].items()
                }
                conf_total += (
                    len(
                        list(
                            act
                            for act in trace_info[i]
                            if act in this_constraints
                        )
                    )
                    + len(
                        list(
                            act
                            for act in trace_info[i]
                            if act not in this_constraints
                        )
                    )
                    + len(
                        list(
                            act
                            for act in this_constraints
                            if min(this_constraints[act]) > 0
                            and act not in trace
                        )
                    )
                )
                for act in trace_info[i]:
                    if act in this_constraints:
                        if trace_info[i][act] not in this_constraints[act]:
                            dev_total += 1
                            ret[Outputs.DEVIATIONS.value].append(
                                (
                                    default_considered_constraints[i],
                                    (act, trace_info[i][act]),
                                )
                            )
                    else:
                        dev_total += 1
                        ret[Outputs.DEVIATIONS.value].append(
                            (default_considered_constraints[i], (act, 0))
                        )
                for act in this_constraints:
                    if min(this_constraints[act]) > 0 and act not in trace:
                        dev_total += 1
                        ret[Outputs.DEVIATIONS.value].append(
                            (default_considered_constraints[i], (act, 0))
                        )
            elif (
                default_considered_constraints[i]
                == DiscoveryOutputs.NEVER_TOGETHER.value
            ):
                this_constraints = {
                    x
                    for x in model[default_considered_constraints[i]]
                    if x[0] in trace
                }
                conf_total += len(this_constraints)
                setinte = this_constraints.intersection(trace_info[i])
                dev_total += len(setinte)
                if len(setinte) > 0:
                    ret[Outputs.DEVIATIONS.value].append(
                        (default_considered_constraints[i], tuple(setinte))
                    )
            else:
                this_constraints = {
                    x
                    for x in model[default_considered_constraints[i]]
                    if x[0] in trace
                }
                conf_total += len(this_constraints)
                setdiff = this_constraints.difference(trace_info[i])
                dev_total += len(setdiff)
                if len(setdiff) > 0:
                    ret[Outputs.DEVIATIONS.value].append(
                        (default_considered_constraints[i], tuple(setdiff))
                    )
        i = i + 1
    ret[Outputs.NO_DEV_TOTAL.value] = dev_total
    ret[Outputs.NO_CONSTR_TOTAL.value] = conf_total
    ret[Outputs.DEV_FITNESS.value] = (
        1.0 - float(dev_total) / float(conf_total) if conf_total > 0 else 1.0
    )
    ret[Outputs.DEVIATIONS.value] = sorted(
        ret[Outputs.DEVIATIONS.value], key=lambda x: (x[0], x[1])
    )
    ret[Outputs.IS_FIT.value] = len(ret[Outputs.DEVIATIONS.value]) == 0
    return ret


def apply_from_variants_list(var_list, model, parameters=None):
    """
    Performs conformance checking using the log skeleton,
    applying it from a list of variants

    Parameters
    --------------
    var_list
        List of variants
    model
        Log skeleton model
    parameters
        Parameters

    Returns
    --------------
    conformance_dictio
        Dictionary containing, for each variant, the result
        of log skeleton checking
    """
    if parameters is None:
        parameters = {}

    conformance_output = {}

    for cv in var_list:
        v = cv[0]
        trace = variants_util.variant_to_trace(v, parameters=parameters)

        conformance_output[v] = apply_trace(
            trace, model, parameters=parameters
        )

    return conformance_output


def after_decode(log_skeleton):
    """
    Prepares the log skeleton after decoding

    Parameters
    --------------
    log_skeleton
        Log skeleton

    Returns
    --------------
    log_skeleton
        Log skeleton (with sets instead of lists)
    """
    log_skeleton[DiscoveryOutputs.EQUIVALENCE.value] = set(
        log_skeleton[DiscoveryOutputs.EQUIVALENCE.value]
    )
    log_skeleton[DiscoveryOutputs.ALWAYS_AFTER.value] = set(
        log_skeleton[DiscoveryOutputs.ALWAYS_AFTER.value]
    )
    log_skeleton[DiscoveryOutputs.ALWAYS_BEFORE.value] = set(
        log_skeleton[DiscoveryOutputs.ALWAYS_BEFORE.value]
    )
    log_skeleton[DiscoveryOutputs.NEVER_TOGETHER.value] = set(
        log_skeleton[DiscoveryOutputs.NEVER_TOGETHER.value]
    )
    log_skeleton[DiscoveryOutputs.DIRECTLY_FOLLOWS.value] = set(
        log_skeleton[DiscoveryOutputs.DIRECTLY_FOLLOWS.value]
    )
    for act in log_skeleton[DiscoveryOutputs.ACTIV_FREQ.value]:
        log_skeleton[DiscoveryOutputs.ACTIV_FREQ.value][act] = set(
            log_skeleton[DiscoveryOutputs.ACTIV_FREQ.value][act]
        )
    return log_skeleton


def get_diagnostics_dataframe(log, conf_result, parameters=None):
    """
    Gets the diagnostics dataframe from a log and the results
    of log skeleton-based conformance checking

    Parameters
    --------------
    log
        Event log
    conf_result
        Results of conformance checking

    Returns
    --------------
    diagn_dataframe
        Diagnostics dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, xes_constants.DEFAULT_TRACEID_KEY
    )

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG, parameters=parameters)

    diagn_stream = []

    for index in range(len(log)):
        case_id = log[index].attributes[case_id_key]

        no_dev_total = conf_result[index][Outputs.NO_DEV_TOTAL.value]
        no_constr_total = conf_result[index][Outputs.NO_CONSTR_TOTAL.value]
        dev_fitness = conf_result[index][Outputs.DEV_FITNESS.value]

        diagn_stream.append(
            {
                "case_id": case_id,
                "no_dev_total": no_dev_total,
                "no_constr_total": no_constr_total,
                "dev_fitness": dev_fitness,
            }
        )

    return pandas_utils.instantiate_dataframe(diagn_stream)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.statistics.attributes.log.get import (
    get_attribute_values,
    get_all_event_attributes_from_log,
    get_all_trace_attributes_from_log,
    get_trace_attribute_values,
)
from pm4py.objects.log.util import sampling
from typing import Union, List, Set
from pm4py.objects.log.obj import EventLog
from pm4py.objects.conversion.log import converter as log_converter


DEFAULT_MAX_CASES_FOR_ATTR_SELECTION = 50


def select_attributes_from_log_for_tree(
    log: EventLog,
    max_cases_for_attr_selection=DEFAULT_MAX_CASES_FOR_ATTR_SELECTION,
    max_diff_occ=DEFAULT_MAX_CASES_FOR_ATTR_SELECTION / 4,
):
    """
    Select attributes from log for tree

    Parameters
    ------------
    log
        Log
    max_cases_for_attr_selection
        Maximum number of cases to consider for attribute selection
    max_diff_occ
        Maximum number of different occurrences

    Returns
    ------------

    """
    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    if len(log) > max_cases_for_attr_selection:
        filtered_log = sampling.sample(log, max_cases_for_attr_selection)
    else:
        filtered_log = log
    event_attributes = get_all_event_attributes_from_log(filtered_log)
    trace_attributes = get_all_trace_attributes_from_log(filtered_log)
    event_attributes_values = {}
    trace_attributes_values = {}
    for attr in event_attributes:
        event_attributes_values[attr] = set(
            get_attribute_values(log, attr).keys()
        )
    for attr in trace_attributes:
        trace_attributes_values[attr] = set(
            get_trace_attribute_values(log, attr).keys()
        )

    numeric_event_attributes_to_consider = list()
    string_event_attributes_to_consider = list()
    numeric_trace_attributes_to_consider = list()
    string_trace_attributes_to_consider = list()

    for attr in event_attributes_values:
        lst = list(event_attributes_values[attr])
        val = lst[0]
        if type(val) is int or type(val) is float:
            numeric_event_attributes_to_consider.append(attr)
        elif type(val) is str and len(lst) < max_diff_occ:
            string_event_attributes_to_consider.append(attr)

    for attr in trace_attributes_values:
        lst = list(trace_attributes_values[attr])
        val = lst[0]
        if type(val) is int or type(val) is float:
            numeric_trace_attributes_to_consider.append(attr)
        elif type(val) is str and len(lst) < max_diff_occ:
            string_trace_attributes_to_consider.append(attr)

    numeric_event_attributes_to_consider = check_event_attributes_presence(
        log, numeric_event_attributes_to_consider
    )
    string_event_attributes_to_consider = check_event_attributes_presence(
        log, string_event_attributes_to_consider
    )
    numeric_trace_attributes_to_consider = check_trace_attributes_presence(
        log, numeric_trace_attributes_to_consider
    )
    string_trace_attributes_to_consider = check_trace_attributes_presence(
        log, string_trace_attributes_to_consider
    )

    return (
        string_trace_attributes_to_consider,
        string_event_attributes_to_consider,
        numeric_trace_attributes_to_consider,
        numeric_event_attributes_to_consider,
    )


def check_trace_attributes_presence(
    log: EventLog, attributes_set: Union[Set[str], List[str]]
) -> Union[Set[str], List[str]]:
    """
    Check trace attributes presence in all the traces of the log

    Parameters
    ------------
    log
        Log
    attributes_set
        Set of attributes

    Returns
    ------------
    filtered_set
        Filtered set of attributes
    """
    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    keys = list(attributes_set)
    for attr in keys:
        if not verify_if_trace_attribute_is_in_each_trace(log, attr):
            attributes_set.remove(attr)
    return attributes_set


def check_event_attributes_presence(
    log: EventLog, attributes_set: Union[Set[str], List[str]]
) -> Union[Set[str], List[str]]:
    """
    Check event attributes presence in all the traces of the log

    Parameters
    ------------
    log
        Log
    attributes_set
        Set of attributes

    Returns
    ------------
    filtered_set
        Filtered set of attributes
    """
    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    keys = list(attributes_set)
    for attr in keys:
        if not verify_if_event_attribute_is_in_each_trace(log, attr):
            attributes_set.remove(attr)
    return attributes_set


def verify_if_event_attribute_is_in_each_trace(
    log: EventLog, attribute: str
) -> bool:
    """
    Verify if the event attribute is in each trace

    Parameters
    ------------
    log
        Log
    attribute
        Attribute

    Returns
    ------------
    boolean
        Boolean value that is aiming to check if the event attribute is in each trace
    """
    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    for trace in log:
        present = False
        for event in trace:
            if attribute in event:
                present = True
                break
        if not present:
            return False
    return True


def verify_if_trace_attribute_is_in_each_trace(
    log: EventLog, attribute: str
) -> bool:
    """
    Verify if the trace attribute is in each trace

    Parameters
    -------------
    log
        Log
    attribute
        Attribute

    Returns
    ------------
    boolean
        Boolean value that is aiming to check if the trace attribute is in each trace
    """
    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    for trace in log:
        if attribute not in trace.attributes:
            return False
    return True



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.visualization.footprints.variants import (
    comparison,
    single,
    comparison_symmetric,
)
from enum import Enum
from pm4py.util import exec_utils, constants
from pm4py.visualization.common import gview
from pm4py.visualization.common import save as gsave
from pm4py.visualization.common.gview import serialize, serialize_dot
import graphviz
from typing import Optional, Dict, Any


class Variants(Enum):
    COMPARISON = comparison
    SINGLE = single
    COMPARISON_SYMMETRIC = comparison_symmetric


def apply(
    *args, variant=None, parameters: Optional[Dict[Any, Any]] = None
) -> graphviz.Source:
    """
    Visualize a footprints table or a comparison between footprints
    tables

    Parameters
    ---------------
    args
        Arguments:
        - A single footprint table
        - Two footprints table (first one associated to the log, second
        one associated to the model)
    parameters
        Parameters of the algorithm, including:
            - Parameters.FORMAT => Format of the visualization

    Returns
    ---------------
    gviz
        Graphviz object
    """
    if parameters is None:
        parameters = {}

    if variant is None:
        if len(args) == 1:
            variant = Variants.SINGLE
        else:
            variant = Variants.COMPARISON

    if variant in [Variants.SINGLE]:
        return exec_utils.get_variant(variant).apply(
            args[0], parameters=parameters
        )
    elif variant in [Variants.COMPARISON, Variants.COMPARISON_SYMMETRIC]:
        return exec_utils.get_variant(variant).apply(
            args[0], args[1], parameters=parameters
        )


def save(gviz, output_file_path, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)
    return ""


def view(gviz, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """
    if constants.DEFAULT_ENABLE_VISUALIZATIONS_VIEW:
        return gview.matplotlib_view(gviz, parameters=parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.petri_net.obj import PetriNet


class Parameters(Enum):
    K = "k"


def apply(
    petri_net: PetriNet,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> float:
    """
    Gets simplicity from a Petri net

    Sánchez-González, Laura, et al. "Prediction of business process model quality based on structural metrics." International Conference on Conceptual Modeling. Berlin, Heidelberg: Springer Berlin Heidelberg, 2010.
    
    Vázquez-Barreiros, Borja, Manuel Mucientes, and Manuel Lama. "ProDiGen: Mining complete, precise and minimal
    structure process models with a genetic algorithm." Information Sciences 294 (2015): 315-333.

    Parameters
    -----------
    petri_net
        Petri net
    parameters
        Possible parameters of the algorithm:
            - K: defines the value to be substracted in the formula: the lower is the value,
            the lower is the simplicity value. k is the baseline arc degree (that is subtracted from the others)

    Returns
    -----------
    simplicity
        Simplicity measure associated to the Petri net
    """
    if parameters is None:
        parameters = {}

    # original model: we have plenty of choices there.
    # one choice is about taking a model containing the most frequent variant,
    # along with a short circuit between the final and the initial marking.
    # in that case, the average arc degree of the "original model" is 2

    # keep the default to 2
    k = exec_utils.get_param_value(Parameters.K, parameters, 2)

    # TODO: verify the real provenence of the approach before!

    all_arc_degrees = []
    for place in petri_net.places:
        all_arc_degrees.append(len(place.in_arcs) + len(place.out_arcs))
    for trans in petri_net.transitions:
        all_arc_degrees.append(len(trans.in_arcs) + len(trans.out_arcs))

    from statistics import mean

    mean_degree = mean(all_arc_degrees) if all_arc_degrees else 0.0

    simplicity = 1.0 / (1.0 + max(mean_degree - k, 0))

    return simplicity



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.petri_net.utils.reachability_graph import (
    construct_reachability_graph,
)
from pm4py.objects.conversion.log import converter as log_converter


def get_tangible_reachability_from_log_net_im_fm(
    log, net, im, fm, parameters=None
):
    """
    Gets the tangible reachability graph from a log and an accepting Petri net

    Parameters
    ---------------
    log
        Event log
    net
        Petri net
    im
        Initial marking
    fm
        Final marking

    Returns
    ------------
    reachab_graph
        Reachability graph
    tangible_reach_graph
        Tangible reachability graph
    stochastic_info
        Stochastic information
    """
    if parameters is None:
        parameters = {}

    from pm4py.algo.simulation.montecarlo.utils import replay

    stochastic_info = replay.get_map_from_log_and_net(
        log_converter.apply(
            log,
            variant=log_converter.Variants.TO_EVENT_LOG,
            parameters=parameters,
        ),
        net,
        im,
        fm,
        parameters=parameters,
    )

    reachability_graph, tangible_reachability_graph = (
        get_tangible_reachability_from_net_im_sinfo(
            net, im, stochastic_info, parameters=parameters
        )
    )

    return reachability_graph, tangible_reachability_graph, stochastic_info


def get_tangible_reachability_from_net_im_sinfo(
    net, im, stochastic_info, parameters=None
):
    """
    Gets the tangible reacahbility graph from a Petri net, an initial marking and a stochastic map

    Parameters
    -------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    stochastic_info
        Stochastic information

    Returns
    ------------
    reachab_graph
        Reachability graph
    tangible_reach_graph
        Tangible reachability graph
    """
    if parameters is None:
        parameters = {}
    reachab_graph = construct_reachability_graph(net, im, use_trans_name=True)
    tang_reach_graph = get_tangible_reachability_from_reachability(
        reachab_graph, stochastic_info
    )

    return reachab_graph, tang_reach_graph


def get_tangible_reachability_from_reachability(reach_graph, stochastic_info):
    """
    Gets the tangible reachability graph from the reachability graph and the stochastic transition map

    Parameters
    ------------
    reach_graph
        Reachability graph
    stochastic_info
        Stochastic information

    Returns
    ------------
    tangible_reach_graph
        Tangible reachability graph
    """
    timed_transitions = []
    for trans in stochastic_info.keys():
        random_variable = stochastic_info[trans]
        transition_type = random_variable.get_transition_type()
        if transition_type == "TIMED":
            timed_transitions.append(trans.name)
    states_reach = list(reach_graph.states)
    for s in states_reach:
        state_outgoing_trans = list(s.outgoing)
        state_ingoing_trans = list(s.incoming)
        timed_trans_outgoing = [
            x for x in state_outgoing_trans if x.name in timed_transitions
        ]
        if not len(state_outgoing_trans) == len(timed_trans_outgoing):
            for t in state_outgoing_trans:
                reach_graph.transitions.remove(t)
                t.from_state.outgoing.remove(t)
                t.to_state.incoming.remove(t)
            for t in state_ingoing_trans:
                reach_graph.transitions.remove(t)
                t.from_state.outgoing.remove(t)
                t.to_state.incoming.remove(t)
            reach_graph.states.remove(s)
    states_reach = list(reach_graph.states)
    for s in states_reach:
        if len(s.incoming) == 0 and len(s.outgoing) == 0:
            reach_graph.states.remove(s)

    return reach_graph



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.util import exec_utils, nx_utils
from pm4py.algo.transformation.ocel.graphs import (
    object_interaction_graph,
    object_descendants_graph,
    object_inheritance_graph,
    object_cobirth_graph,
    object_codeath_graph,
)


class Parameters(Enum):
    INCLUDE_OBJ_INTERACTION_GRAPH = "include_obj_interaction_graph"
    INCLUDE_OBJ_DESCENDANTS_GRAPH = "include_obj_descendants_graph"
    INCLUDE_OBJ_INHERITANCE_GRAPH = "include_obj_inheritance_graph"
    INCLUDE_OBJ_COBIRTH_GRAPH = "include_obj_cobirth_graph"
    INCLUDE_OBJ_CODEATH_GRAPH = "include_obj_codeath_graph"


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Converts the graphs of features extracted from an OCEL to a NetworkX DiGraph object

    Parameters
    --------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.INCLUDE_OBJ_INTERACTION_GRAPH => includes the object interaction graph in the NX DiGraph
        - Parameters.INCLUDE_OBJ_DESCENDANTS_GRAPH => includes the object descendants graph in the NX DiGraph
        - Parameters.INCLUDE_OBJ_INHERITANCE_GRAPH => includes the object inheritance graph in the NX DiGraph
        - Parameters.INCLUDE_OBJ_COBIRTH_GRAPH => includes the object cobirth graph in the NX DiGraph
        - Parameters.INCLUDE_OBJ_CODEATH_GRAPH => includes the object codeath graph in the NX DiGraph

    Returns
    -------------
    G
        NetworkX DiGraph
    """
    if parameters is None:
        parameters = {}

    include_obj_interaction_graph = exec_utils.get_param_value(
        Parameters.INCLUDE_OBJ_INTERACTION_GRAPH, parameters, True
    )
    include_obj_descendants_graph = exec_utils.get_param_value(
        Parameters.INCLUDE_OBJ_DESCENDANTS_GRAPH, parameters, True
    )
    include_obj_inheritance_graph = exec_utils.get_param_value(
        Parameters.INCLUDE_OBJ_INHERITANCE_GRAPH, parameters, True
    )
    include_obj_cobirth_graph = exec_utils.get_param_value(
        Parameters.INCLUDE_OBJ_COBIRTH_GRAPH, parameters, True
    )
    include_obj_codeath_graph = exec_utils.get_param_value(
        Parameters.INCLUDE_OBJ_CODEATH_GRAPH, parameters, True
    )

    G = nx_utils.DiGraph()
    if include_obj_interaction_graph:
        interaction_graph = object_interaction_graph.apply(
            ocel, parameters=parameters
        )
        G.add_edges_from(interaction_graph, attr={"type": "INTERACTION"})

    if include_obj_descendants_graph:
        descendants_graph = object_descendants_graph.apply(
            ocel, parameters=parameters
        )
        G.add_edges_from(descendants_graph, attr={"type": "DESCENDANTS"})

    if include_obj_inheritance_graph:
        inheritance_graph = object_inheritance_graph.apply(
            ocel, parameters=parameters
        )
        G.add_edges_from(inheritance_graph, attr={"type": "INHERITANCE"})

    if include_obj_cobirth_graph:
        cobirth_graph = object_cobirth_graph.apply(ocel, parameters=parameters)
        G.add_edges_from(cobirth_graph, attr={"type": "COBIRTH"})

    if include_obj_codeath_graph:
        codeath_graph = object_codeath_graph.apply(ocel, parameters=parameters)
        G.add_edges_from(codeath_graph, attr={"type": "CODEATH"})

    return G



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from pm4py.util import nx_utils


def compute_incidence_matrix(net):
    """
    Given a Petri Net, the incidence matrix is computed. An incidence matrix has n rows (places) and m columns
    (transitions).
    :param net: Petri Net object
    :return: Incidence matrix
    """
    n = len(net.transitions)
    m = len(net.places)
    C = np.zeros((m, n))
    i = 0
    transition_list = sorted(list(net.transitions), key=lambda x: x.name)
    place_list = sorted(list(net.places), key=lambda x: x.name)
    while i < n:
        t = transition_list[i]
        for in_arc in t.in_arcs:
            # arcs that go to transition
            C[place_list.index(in_arc.source), i] -= 1 * in_arc.weight
        for out_arc in t.out_arcs:
            # arcs that lead away from transition
            C[place_list.index(out_arc.target), i] += 1 * out_arc.weight
        i += 1
    return C


def split_incidence_matrix(matrix, net):
    """
    We split the incidence matrix columnwise to get the firing information for each transition
    :param matrix: incidence matrix
    :param net: Petri Net
    :return: Dictionary, whereby the key is an np array that contains the firing information and the value is the name
    of the transition
    """
    transition_dict = {}
    lst_transitions = sorted(list(net.transitions), key=lambda x: x.name)
    i = 0
    while i < len(net.transitions):
        transition_dict[lst_transitions[i]] = np.hsplit(
            np.transpose(matrix), 1
        )[0][i]
        i += 1
    return transition_dict


def compute_firing_requirement(net):
    place_list = sorted(list(net.places), key=lambda x: x.name)
    transition_dict = {}
    for transition in net.transitions:
        temp_array = np.zeros(len(place_list))
        for arc in transition.in_arcs:
            temp_array[place_list.index(arc.source)] -= 1 * arc.weight
        transition_dict[transition] = temp_array
    return transition_dict


def enabled_markings(firing_dict, req_dict, marking):
    enabled_transitions = []
    for transition, requirment in req_dict.items():
        if all(np.greater_equal(marking, requirment.copy() * -1)):
            enabled_transitions.append(transition)
    new_markings = []
    for transition in enabled_transitions:
        new_marking = marking + firing_dict[transition]
        new_markings.append((new_marking, transition))
    return new_markings


def convert_marking(net, marking, original_net=None):
    """
    Takes an marking as input and converts it into an Numpy Array
    :param net: PM4Py Petri Net object
    :param marking: Marking that should be converted
    :param original_net: PM4Py Petri Net object without short-circuited transition
    :return: Numpy array representation
    """
    # marking_list=list(el.name for el in marking.keys())
    #
    marking_list = sorted([el.name for el in marking.keys()])
    place_list = sorted(list(el.name for el in net.places))

    mark = np.zeros(len(place_list))
    for index, value in enumerate(mark):
        if place_list[index] in marking_list:
            # TODO: Is setting the value to 1 ok in this case?
            mark[index] = 1
    return mark


def check_for_dead_tasks(net, graph):
    """
    We compute a list of dead tasks. A dead task is a task which does not appear in the Minimal Coverability Graph
    :param net: Petri Net representation of PM4Py
    :param graph: Minimal coverability graph. NetworkX MultiDiGraph object.
    :return: list of dead tasks
    """
    tasks = []
    lst_transitions = sorted(list(net.transitions), key=lambda x: x.name)
    for transition in lst_transitions:
        if transition.label is not None:
            tasks.append(transition)
    for node, targets in graph.edges()._adjdict.items():
        for target_node, activties in targets.items():
            for option, activity in activties.items():
                if activity["transition"] in tasks:
                    tasks.remove(activity["transition"])
    return tasks


def check_for_improper_conditions(mcg):
    """
    An improper condition is a state in the minimum-coverability graph with an possible infinite amount of tokens
    :param mcg: networkx object (minimal coverability graph)
    :return: True, if there are no improper conditions; false otherwise
    """
    improper_states = []
    for node in mcg.nodes:
        if np.inf in mcg.nodes[node]["marking"]:
            improper_states.append(node)
    return improper_states


def check_for_substates(mcg):
    """
    Checks if a substate exists in a given mcg
    :param mcg: Minimal coverability graph (networkx object)
    :return: True, if there exist no substate; False otherwise
    """
    for node in mcg.nodes:
        reachable_states = nx_utils.descendants(mcg, node)
        for state in reachable_states:
            if all(
                np.less(
                    mcg.nodes[node]["marking"], mcg.nodes[state]["marking"]
                )
            ):
                return False
    return True



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.util import pandas_utils
from pm4py.algo.transformation.ocel.features.objects import (
    object_lifecycle_duration,
)
import pandas as pd


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Assigns to each object a feature which is the number of objects that are open during the lifecycle
    of the object.

    Parameters
    ----------------
    ocel
        OCEL
    parameters
        Parameters of the algorithm:
            - Parameters.OBJECT_STR_ATTRIBUTES => collection of string attributes to consider for feature extraction.

    Returns
    ----------------
    data
        Extracted feature values
    feature_names
        Feature names
    """
    if parameters is None:
        parameters = {}

    data, feature_names = object_lifecycle_duration.apply(
        ocel, parameters=parameters
    )
    obj_dur = pandas_utils.instantiate_dataframe(data, columns=feature_names)

    obj_dur["@@index"] = obj_dur.index
    obj_dur = obj_dur.to_dict("records")

    data = []
    obj_dur.sort(
        key=lambda x: (
            x["@@object_lifecycle_start_timestamp"],
            x["@@object_lifecycle_end_timestamp"],
        )
    )
    for i in range(len(obj_dur)):
        j = i + 1
        ct = obj_dur[i]["@@object_lifecycle_end_timestamp"]
        while j < len(obj_dur):
            st = obj_dur[j]["@@object_lifecycle_start_timestamp"]
            if st > ct:
                break
            j = j + 1
        data.append([j - i, obj_dur[i]["@@index"]])
    feature_names = ["@@object_wip"]

    data.sort(key=lambda x: x[1])
    data = [[float(x[0])] for x in data]

    return data, feature_names



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.objects.ocel.util import filtering_utils
from pm4py.objects.ocel.util import ocel_consistency
from pm4py.objects.ocel.importer.jsonocel.variants import classic
from pm4py.util import constants as pm4_constants
from enum import Enum
import json


class Parameters(Enum):
    ENCODING = "encoding"


def apply(file_path: str, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    """
    Imports an OCEL from a JSON-OCEL 2 standard file

    Parameters
    --------------
    file_path
        Path to the object-centric event log
    parameters
        Possible parameters of the method, including:
        - Parameters.ENCODING

    Returns
    -------------
    ocel
        Object-centric event log
    """
    if parameters is None:
        parameters = {}

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, pm4_constants.DEFAULT_ENCODING
    )

    F = open(file_path, "r", encoding=encoding)
    json_obj = json.load(F)
    F.close()

    legacy_obj = {}
    legacy_obj["ocel:events"] = {}
    legacy_obj["ocel:objects"] = {}
    legacy_obj["ocel:objectChanges"] = []

    for eve in json_obj["events"]:
        dct = {}
        dct["ocel:activity"] = eve["type"]
        dct["ocel:timestamp"] = eve["time"]
        dct["ocel:vmap"] = {}
        if "attributes" in eve and eve["attributes"]:
            dct["ocel:vmap"] = {
                x["name"]: x["value"] for x in eve["attributes"]
            }
        dct["ocel:typedOmap"] = []
        if "relationships" in eve and eve["relationships"]:
            dct["ocel:typedOmap"] = [
                {"ocel:oid": x["objectId"], "ocel:qualifier": x["qualifier"]}
                for x in eve["relationships"]
            ]
        dct["ocel:omap"] = list(
            set(x["ocel:oid"] for x in dct["ocel:typedOmap"])
        )
        legacy_obj["ocel:events"][eve["id"]] = dct

    for obj in json_obj["objects"]:
        dct = {}
        dct["ocel:type"] = obj["type"]
        dct["ocel:ovmap"] = {}
        if "attributes" in obj and obj["attributes"]:
            for x in obj["attributes"]:
                if x["name"] in dct["ocel:ovmap"]:
                    legacy_obj["ocel:objectChanges"].append(
                        {
                            "ocel:oid": obj["id"],
                            "ocel:type": obj["type"],
                            "ocel:field": x["name"],
                            x["name"]: x["value"],
                            "ocel:timestamp": x["time"],
                        }
                    )
                else:
                    dct["ocel:ovmap"][x["name"]] = x["value"]
        dct["ocel:o2o"] = []
        if "relationships" in obj and obj["relationships"]:
            dct["ocel:o2o"] = [
                {"ocel:oid": x["objectId"], "ocel:qualifier": x["qualifier"]}
                for x in obj["relationships"]
            ]
        legacy_obj["ocel:objects"][obj["id"]] = dct

    legacy_obj["ocel:global-log"] = {}
    legacy_obj["ocel:global-event"] = {}
    legacy_obj["ocel:global-object"] = {}

    log = classic.get_base_ocel(legacy_obj, parameters=parameters)

    log = ocel_consistency.apply(log, parameters=parameters)
    log = filtering_utils.propagate_relations_filtering(
        log, parameters=parameters
    )

    return log



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import numpy as np
import warnings
from scipy.optimize import linprog, OptimizeResult
from typing import Optional, Dict, Any, List
from pm4py.util import exec_utils
from threading import Lock
from scipy.optimize import OptimizeWarning


warnings.filterwarnings('ignore', category=OptimizeWarning)
LP_LOCK = Lock()


class Parameters:
    INTEGRALITY = "integrality"
    METHOD = "method"
    BOUNDS = "bounds"


def apply(
        c: list,
        Aub: np.ndarray,
        bub: np.matrix,
        Aeq: np.matrix,
        beq: np.matrix,
        parameters: Optional[Dict[Any, Any]] = None,
) -> OptimizeResult:
    if parameters is None:
        parameters = {}

    integrality = exec_utils.get_param_value(
        Parameters.INTEGRALITY, parameters, None
    )
    method = exec_utils.get_param_value(Parameters.METHOD, parameters, "revised simplex")
    bounds = exec_utils.get_param_value(Parameters.BOUNDS, parameters, None)

    with LP_LOCK:
        sol = linprog(
            c,
            A_ub=Aub,
            b_ub=bub,
            A_eq=Aeq,
            b_eq=beq,
            integrality=integrality,
            bounds=bounds,
            method=method,
        )

    return sol


def get_prim_obj_from_sol(
        sol: OptimizeResult, parameters: Optional[Dict[Any, Any]] = None
) -> Optional[int]:
    if sol is not None and sol.fun is not None:
        return round(sol.fun)
    return None


def get_points_from_sol(
        sol: OptimizeResult, parameters: Optional[Dict[Any, Any]] = None
) -> Optional[List[int]]:
    if sol is not None and sol.x is not None:
        return [round(y) for y in sol.x]
    return None



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Optional, Dict, Any, List

from pm4py.objects.ocel import constants
from pm4py.objects.ocel.obj import OCEL


def get_attribute_names(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> List[str]:
    """
    Gets the list of attributes at the event and the object level of an object-centric event log
    (e.g. ["cost", "amount", "name"])

    Parameters
    -------------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm

    Returns
    -------------------
    attributes_list
        List of attributes at the event and object level (e.g. ["cost", "amount", "name"])
    """
    if parameters is None:
        parameters = {}

    attributes = sorted(
        set(
            x
            for x in ocel.events.columns
            if not x.startswith(constants.OCEL_PREFIX)
        ).union(
            x
            for x in ocel.objects.columns
            if not x.startswith(constants.OCEL_PREFIX)
        )
    )

    return attributes



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util.constants import CASE_CONCEPT_NAME
from pm4py.util import xes_constants as xes
from pm4py.util.xes_constants import DEFAULT_NAME_KEY
from pm4py.util.constants import (
    PARAMETER_CONSTANT_CASEID_KEY,
    PARAMETER_CONSTANT_ACTIVITY_KEY,
    GROUPED_DATAFRAME,
    RETURN_EA_COUNT_DICT_AUTOFILTER,
)
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List
import polars as pl


class Parameters(Enum):
    CASE_ID_KEY = PARAMETER_CONSTANT_CASEID_KEY
    ACTIVITY_KEY = PARAMETER_CONSTANT_ACTIVITY_KEY
    DECREASING_FACTOR = "decreasingFactor"
    GROUP_DATAFRAME = GROUPED_DATAFRAME
    POSITIVE = "positive"
    RETURN_EA_COUNT = RETURN_EA_COUNT_DICT_AUTOFILTER


def get_end_activities(
    df: pl.LazyFrame,
    case_id_key: str = CASE_CONCEPT_NAME,
    activity_key: str = DEFAULT_NAME_KEY,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, int]:
    """
    Get end activities from a LazyFrame with their counts

    Parameters
    ----------
    df
        LazyFrame
    case_id_key
        Case ID column
    activity_key
        Activity column
    parameters
        Parameters

    Returns
    ----------
    end_activities_dict
        Dictionary of end activities and their counts
    """
    if parameters is None:
        parameters = {}
    
    # Get last activity for each case
    end_activities = (
        df.group_by(case_id_key)
        .agg(pl.col(activity_key).last().alias("end_activity"))
        .group_by("end_activity")
        .count()
        .collect()
    )
    
    # Convert to dictionary
    result = {}
    for row in end_activities.iter_rows():
        result[row[0]] = row[1]
    
    return result


def apply(
    df: pl.LazyFrame,
    values: List[str],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on end activities

    Parameters
    ----------
    df
        LazyFrame
    values
        Values to filter on
    parameters
        Possible parameters of the algorithm, including:
            Parameters.CASE_ID_KEY -> Case ID column in the LazyFrame
            Parameters.ACTIVITY_KEY -> Column that represents the activity
            Parameters.POSITIVE -> Specifies if the filtered should be applied including traces (positive=True)
            or excluding traces (positive=False)

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    if parameters is None:
        parameters = {}

    case_id_glue = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, CASE_CONCEPT_NAME
    )
    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, DEFAULT_NAME_KEY
    )
    positive = exec_utils.get_param_value(
        Parameters.POSITIVE, parameters, True
    )

    return filter_df_on_end_activities(
        df,
        values,
        case_id_glue=case_id_glue,
        activity_key=activity_key,
        positive=positive,
    )


def filter_df_on_end_activities(
    df: pl.LazyFrame,
    values: List[str],
    case_id_glue: str = CASE_CONCEPT_NAME,
    activity_key: str = xes.DEFAULT_NAME_KEY,
    positive: bool = True,
) -> pl.LazyFrame:
    """
    Filter LazyFrame on end activities

    Parameters
    ----------
    df
        LazyFrame
    values
        Values to filter on
    case_id_glue
        Case ID column in the LazyFrame
    activity_key
        Column that represent the activity
    positive
        Specifies if the filtered should be applied including traces (positive=True) or excluding traces
        (positive=False)

    Returns
    ----------
    df
        Filtered LazyFrame
    """
    # Get cases with matching end activities
    matching_cases = (
        df.group_by(case_id_glue)
        .agg(pl.col(activity_key).last().alias("end_activity"))
        .filter(pl.col("end_activity").is_in(values))
        .select(case_id_glue)
    )
    
    if positive:
        # Keep cases with matching end activities
        ret = df.join(matching_cases, on=case_id_glue, how="inner")
    else:
        # Keep cases without matching end activities
        ret = df.join(matching_cases, on=case_id_glue, how="anti")

    return ret


def filter_df_on_end_activities_nocc(
    df: pl.LazyFrame,
    nocc: int,
    ea_count0: Optional[Dict[str, int]] = None,
    case_id_glue: str = CASE_CONCEPT_NAME,
    activity_key: str = xes.DEFAULT_NAME_KEY,
    return_dict: bool = False,
    most_common_variant: Optional[List[str]] = None,
):
    """
    Filter LazyFrame on end activities number of occurrences

    Parameters
    -----------
    df
        LazyFrame
    nocc
        Minimum number of occurrences of the end activity
    ea_count0
        (if provided) Dictionary that associates each end activity with its count
    case_id_glue
        Column that contains the Case ID
    activity_key
        Column that contains the activity
    return_dict
        Return dict
    most_common_variant
        Most common variant (if provided)
    """
    if most_common_variant is None:
        most_common_variant = []

    if ea_count0 is None:
        ea_count0 = get_end_activities(df, case_id_glue, activity_key)
    
    ea_count = [
        k
        for k, v in ea_count0.items()
        if v >= nocc
        or (len(most_common_variant) > 0 and k == most_common_variant[-1])
    ]
    ea_count_dict = {
        k: v
        for k, v in ea_count0.items()
        if v >= nocc
        or (len(most_common_variant) > 0 and k == most_common_variant[-1])
    }
    
    if len(ea_count) < len(ea_count0):
        ret = filter_df_on_end_activities(
            df, ea_count, case_id_glue=case_id_glue, activity_key=activity_key
        )
        if return_dict:
            return ret, ea_count_dict
        return ret
    
    if return_dict:
        return df, ea_count_dict
    return df